"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictBool, StrictStr, field_validator

from snowflake.core.cortex.analyst_service._generated.models.message_object import MessageObject, MessageObjectModel
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_object import (
    SemanticModelObject,
    SemanticModelObjectModel,
)


class SendMessageRequest(BaseModel):
    """A model object representing the SendMessageRequest resource.

    Constructs an object of type SendMessageRequest with the provided properties.

    Parameters
    __________
    messages : list[MessageObject]

    semantic_model_file : str, optional
        The path to a file stored in a Snowflake Stage holding the semantic model yaml. Must be a fully qualified stage url
    semantic_model : str, optional
        A string containing the entire semantic model yaml
    semantic_view : str, optional
        The name of the Snowflake native semantic model object
    semantic_models : list[SemanticModelObject], optional
        A list of semantic model objects. If set, other semantic model properties are ignored
    stream : bool,  default False
        Whether to stream the response or not
    operation : str,  default 'sql_generation'
        Whether to response with SQL or natural language. One of 'sql_generation' or 'answer_generation'
    warehouse : str, optional
        Warehouse name to use for result set handling. Only used when 'operation' is 'answer_generation'
    source : str, optional
        an optional field to specify the source of the request. e.g "eval", "prod"
    experimental : str, optional
        JSON serialized string of experimental API fields (undocumented).
    """

    semantic_model_file: Optional[StrictStr] = None

    semantic_model: Optional[StrictStr] = None

    semantic_view: Optional[StrictStr] = None

    semantic_models: Optional[List[SemanticModelObject]] = None

    stream: Optional[StrictBool] = False

    operation: Optional[StrictStr] = "sql_generation"

    warehouse: Optional[StrictStr] = None

    messages: List[MessageObject]

    source: Optional[StrictStr] = None

    experimental: Optional[StrictStr] = None

    __properties = [
        "semantic_model_file",
        "semantic_model",
        "semantic_view",
        "semantic_models",
        "stream",
        "operation",
        "warehouse",
        "messages",
        "source",
        "experimental",
    ]

    @field_validator("operation")
    def operation_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("sql_generation", "answer_generation"):
            raise ValueError("must validate the enum values ('sql_generation','answer_generation')")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SendMessageRequest:
        """Create an instance of SendMessageRequest from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in semantic_models (list)
        _items = []
        if self.semantic_models:
            for _item in self.semantic_models:
                if _item:
                    _items.append(_item.to_dict())
            _dict["semantic_models"] = _items

        # override the default output from pydantic by calling `to_dict()` of each item in messages (list)
        _items = []
        if self.messages:
            for _item in self.messages:
                if _item:
                    _items.append(_item.to_dict())
            _dict["messages"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> SendMessageRequest:
        """Create an instance of SendMessageRequest from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return SendMessageRequest.model_validate(obj)

        _obj = SendMessageRequest.model_validate(
            {
                "semantic_model_file": obj.get("semantic_model_file"),
                "semantic_model": obj.get("semantic_model"),
                "semantic_view": obj.get("semantic_view"),
                "semantic_models": [SemanticModelObject.from_dict(_item) for _item in obj.get("semantic_models")]
                if obj.get("semantic_models") is not None
                else None,
                "stream": obj.get("stream") if obj.get("stream") is not None else False,
                "operation": obj.get("operation") if obj.get("operation") is not None else "sql_generation",
                "warehouse": obj.get("warehouse"),
                "messages": [MessageObject.from_dict(_item) for _item in obj.get("messages")]
                if obj.get("messages") is not None
                else None,
                "source": obj.get("source"),
                "experimental": obj.get("experimental"),
            }
        )

        return _obj


class SendMessageRequestModel:
    def __init__(
        self,
        messages: list[MessageObject],
        # optional properties
        semantic_model_file: Optional[str] = None,
        semantic_model: Optional[str] = None,
        semantic_view: Optional[str] = None,
        semantic_models: Optional[list[SemanticModelObject]] = None,
        stream: Optional[bool] = False,
        operation: Optional[str] = "sql_generation",
        warehouse: Optional[str] = None,
        source: Optional[str] = None,
        experimental: Optional[str] = None,
    ):
        """A model object representing the SendMessageRequest resource.

        Constructs an object of type SendMessageRequest with the provided properties.

        Parameters
        __________
        messages : list[MessageObject]

        semantic_model_file : str, optional
            The path to a file stored in a Snowflake Stage holding the semantic model yaml. Must be a fully qualified stage url
        semantic_model : str, optional
            A string containing the entire semantic model yaml
        semantic_view : str, optional
            The name of the Snowflake native semantic model object
        semantic_models : list[SemanticModelObject], optional
            A list of semantic model objects. If set, other semantic model properties are ignored
        stream : bool,  default False
            Whether to stream the response or not
        operation : str,  default 'sql_generation'
            Whether to response with SQL or natural language. One of 'sql_generation' or 'answer_generation'
        warehouse : str, optional
            Warehouse name to use for result set handling. Only used when 'operation' is 'answer_generation'
        source : str, optional
            an optional field to specify the source of the request. e.g "eval", "prod"
        experimental : str, optional
            JSON serialized string of experimental API fields (undocumented).
        """
        self.semantic_model_file = semantic_model_file
        self.semantic_model = semantic_model
        self.semantic_view = semantic_view
        self.semantic_models = semantic_models
        self.stream = stream
        self.operation = operation
        self.warehouse = warehouse
        self.messages = messages
        self.source = source
        self.experimental = experimental

    __properties = [
        "semantic_model_file",
        "semantic_model",
        "semantic_view",
        "semantic_models",
        "stream",
        "operation",
        "warehouse",
        "messages",
        "source",
        "experimental",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return SendMessageRequest(
            semantic_model_file=self.semantic_model_file,
            semantic_model=self.semantic_model,
            semantic_view=self.semantic_view,
            semantic_models=[x._to_model() for x in self.semantic_models] if self.semantic_models is not None else None,
            stream=self.stream,
            operation=self.operation,
            warehouse=self.warehouse,
            messages=[x._to_model() for x in self.messages] if self.messages is not None else None,
            source=self.source,
            experimental=self.experimental,
        )

    @classmethod
    def _from_model(cls, model) -> SendMessageRequestModel:
        return SendMessageRequestModel(
            semantic_model_file=model.semantic_model_file,
            semantic_model=model.semantic_model,
            semantic_view=model.semantic_view,
            semantic_models=[SemanticModelObjectModel._from_model(x) for x in model.semantic_models]
            if model.semantic_models
            else None,
            stream=model.stream,
            operation=model.operation,
            warehouse=model.warehouse,
            messages=[MessageObjectModel._from_model(x) for x in model.messages] if model.messages else None,
            source=model.source,
            experimental=model.experimental,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a SendMessageRequest.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the SendMessageRequest object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> SendMessageRequestModel:
        """Create an instance of SendMessageRequest from a dict.

        This method constructs a SendMessageRequest object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        SendMessageRequest
            A SendMessageRequest object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(SendMessageRequest.from_dict(obj))


SendMessageRequest._model_class = SendMessageRequestModel



from unittest.mock import MagicMock

import pytest

from snowflake.core._identifiers import FQN
from snowflake.core.exceptions import InvalidIdentifierError


def test_attributes():
    fqn = FQN(name="object_name", database="database_name", schema="schema_name")
    assert fqn.database == "database_name"
    assert fqn.schema == "schema_name"
    assert fqn.name == "object_name"


@pytest.mark.parametrize(
    "fqn, identifier",
    [
        (FQN(name="name", database="db", schema="schema"), "db.schema.name"),
        (FQN(name='"my_object"', database="db", schema="schema"), 'db.schema."my_object"'),
        (FQN(name="name", database=None, schema="schema"), "schema.name"),
        (FQN(name="name", database=None, schema=None), "name"),
        (FQN(name="name", database="db", schema=None), "db.PUBLIC.name"),
        (FQN(name="name(float, string)", database="db", schema=None), "db.PUBLIC.name(float, string)"),
    ],
)
def test_identifier(fqn, identifier):
    assert fqn.identifier == identifier


def test_eq():
    fqn1 = FQN(name="name", database="db", schema="schema")
    fqn2 = FQN(name="name", database="db", schema="schema")
    assert fqn1 == fqn2
    assert fqn2 == fqn1


def test_neq():
    fqn1 = FQN(name="name", database="db", schema="schema")
    fqn2 = FQN(name="other-name", database="db", schema="schema")
    assert fqn1 != fqn2
    assert fqn2 != fqn1


def test_eq_with_different_type():
    assert FQN(name="name", database="db", schema="schema") != "db.schema.name"


def test_set_database():
    fqn = FQN(name="name", database="db", schema="schema")
    fqn.set_database("foo")
    assert fqn.database == "foo"


def test_set_schema():
    fqn = FQN(name="name", database="db", schema="schema")
    fqn.set_schema("foo")
    assert fqn.schema == "foo"


def test_set_name():
    fqn = FQN(name="name", database="db", schema="schema")
    fqn.set_name("foo")
    assert fqn.name == "foo"


@pytest.mark.parametrize(
    "fqn_str, identifier",
    [
        ("db.schema.name", "db.schema.name"),
        ("DB.SCHEMA.NAME", "DB.SCHEMA.NAME"),
        ("schema.name", "schema.name"),
        ("name", "name"),
        ('"name with space"', '"name with space"'),
        ('"dot.db"."dot.schema"."dot.name"', '"dot.db"."dot.schema"."dot.name"'),
        ('"dot.db".schema."dot.name"', '"dot.db".schema."dot.name"'),
        ('db.schema."dot.name"', 'db.schema."dot.name"'),
        ('"dot.db".schema."DOT.name"', '"dot.db".schema."DOT.name"'),
        # Nested quotes
        ('"abc""this is in nested quotes"""', '"abc""this is in nested quotes"""'),
        # Callables
        ("db.schema.function(string, int, variant)", "db.schema.function"),
        ('db.schema."fun tion"(string, int, variant)', 'db.schema."fun tion"'),
    ],
)
def test_from_string(fqn_str, identifier):
    fqn = FQN.from_string(fqn_str)
    assert fqn.identifier == identifier
    if fqn.signature:
        assert fqn.signature == "(string, int, variant)"


@pytest.mark.parametrize(
    "fqn_str",
    [
        "db.schema.name.foo",
        "schema. name",
        "name with space",
        'dot.db."dot.schema"."dot.name"',
        '"dot.db.schema."dot.name"',
    ],
)
def test_from_string_fails_if_pattern_does_not_match(fqn_str):
    with pytest.raises(InvalidIdentifierError) as err:
        FQN.from_string(fqn_str)

    assert str(err.value) == f"'{fqn_str}' is not a valid identifier."


def test_using_connection():
    connection = MagicMock(database="database_test", schema="test_schema")
    fqn = FQN.from_string("name").using_connection(connection)
    assert fqn.identifier == "database_test.test_schema.name"


def test_to_dict():
    fqn = FQN(name="name", database="db", schema="schema")
    assert fqn.to_dict() == {"name": "name", "database": "db", "schema": "schema"}



from typing import TYPE_CHECKING

from pydantic import StrictStr

from snowflake.core._operation import PollingOperation, PollingOperations

from .._internal.telemetry import api_telemetry
from . import ApiIntegration
from ._generated.api.api_integration_api_base import ApiIntegrationCollectionBase, ApiIntegrationResourceBase


if TYPE_CHECKING:
    from snowflake.core import PollingOperation, Root


class ApiIntegrationCollection(ApiIntegrationCollectionBase):
    """Represents the collection operations on the Snowflake api integration resource.

    With this collection, you can create, iterate through, and search for api integration that you
    have access to in the current context.

    Examples
    ________
    Creating an ApiIntegration instance using AWS API Gateway:

    >>> api_integrations = root.api_integrations
    >>> new_api_integration = ApiIntegration(
    ...     name="name",
    ...     api_hook=AwsHook(
    ...         api_provider="AWS_API_GATEWAY",
    ...         api_aws_role_arn="your_arn",
    ...         api_key=os.environ.get("YOUR_API_KEY"),
    ...     ),
    ...     api_allowed_prefixes=["https://snowflake.com"],
    ...     enabled=True,
    ... )
    >>> api_integrations.create(new_api_integration)
    """

    def __init__(self, root: "Root"):
        super().__init__(root, ApiIntegrationResource)


class ApiIntegrationResource(ApiIntegrationResourceBase):
    """Represents a reference to a Snowflake api integration.

    With this api integration reference, you can fetch information about an api integration,
    as well as perform certain actions on it.
    """

    _plural_name = "api_integrations"

    def __init__(self, name: StrictStr, collection: ApiIntegrationCollection) -> None:
        super().__init__(name, collection)

    @api_telemetry
    def create_or_alter(self, api_integration: ApiIntegration) -> None:
        """Create or alter an API integration.

        The operation is limited by the fact that api_key will not be updated
        and api_blocked_prefixes cannot be unset.

        Parameters
        __________
        api_integration: ApiIntegration
            The ``ApiIntegration`` object.

        Examples
        ________
        Creating a new API integration:

        >>> root.api_integrations["my_api"].create_or_alter(my_api_def)

        See ``ApiIntegrationCollection.create`` for more examples.
        """
        self.collection._api.create_or_alter_api_integration(
            api_integration.name, api_integration=api_integration, async_req=False
        )

    @api_telemetry
    def create_or_alter_async(self, api_integration: ApiIntegration) -> PollingOperation[None]:
        """An asynchronous version of :func:`create_or_alter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.create_or_alter_api_integration(
            api_integration.name, api_integration=api_integration, async_req=True
        )
        return PollingOperations.empty(future)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
from textwrap import dedent

import pytest

from snowflake.core.service import Service, ServiceSpecInlineText

from ..utils import random_string


pytestmark = [pytest.mark.skip_gov]


def test_stop_all_services(compute_pools, services, imagerepo, temp_cp):
    compute_pools[temp_cp.name].resume()

    s_name = random_string(5, "test_service_")
    inline_spec = dedent(
        f"""
            spec:
              containers:
              - name: hello-world
                image: {imagerepo}/hello-world:latest
             """
    )
    test_s = Service(
        name=s_name,
        compute_pool=temp_cp.name,
        spec=ServiceSpecInlineText(spec_text=inline_spec),
        min_instances=1,
        max_instances=1,
    )
    services.create(test_s)

    services_in_pool = [service for service in services.iter() if service.compute_pool == temp_cp.name.upper()]
    assert len(services_in_pool) == 1
    assert services_in_pool[0].name == test_s.name.upper()
    compute_pools[temp_cp.name].stop_all_services()
    services_in_pool = [service for service in services.iter() if service.compute_pool == temp_cp.name.upper()]
    assert len(services_in_pool) == 0



import pytest

from snowflake.core.database import Database
from snowflake.core.database_role import DatabaseRole
from tests.integ.utils import backup_database_and_schema
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


def test_iter(databases, cursor):
    try:
        with backup_database_and_schema(cursor):
            new_db_def = Database(name=random_string(3, "test_database_$12create_"), kind="TRANSIENT")
            new_db_def.comment = "database first"
            database_handle = databases.create(new_db_def)
            cursor.execute(f"USE DATABASE {new_db_def.name}")

            database_role_name = random_string(10, "test_database_role")
            database_role = DatabaseRole(name=database_role_name, comment="test_comment")

            database_role_handles = []

            database_roles = databases[new_db_def.name].database_roles

            try:
                # check there are no database roles yet
                assert len(list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))) == 0

                # happy path
                database_role_handles += [database_roles.create(database_role)]
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name.upper()[:-1]))
                assert len(database_roles_list) == 1

                database_role_name_2 = random_string(10, "test_database_role_2")
                database_role_2 = DatabaseRole(name=database_role_name_2, comment="test_comment")
                database_role_handles += [database_roles.create(database_role_2)]
                database_roles_list = list(database_roles.iter(limit=1, from_name=database_role_name_2.upper()[:-1]))
                assert len(database_roles_list) == 1

                # check both roles
                database_roles_list = list(database_roles.iter())
                assert len(database_roles_list) == 2

                # check limit
                database_roles_list = list(database_roles.iter(limit=1))
                assert len(database_roles_list) == 1
            finally:
                for database_role_handle in database_role_handles:
                    database_role_handle.drop()
    finally:
        database_handle.drop()



import typing

from contextlib import suppress
from datetime import timedelta

import pytest

from snowflake.core._internal.utils import normalize_name
from snowflake.core.task import Cron, StoredProcedureCall, Task

from ..utils import random_object_name


if typing.TYPE_CHECKING:
    pass


@pytest.mark.snowpark
@pytest.mark.min_sf_ver("9.4.0")
def test_create_or_alter(tasks, db_parameters):
    from snowflake.snowpark._internal.utils import parse_table_name

    task_name1 = random_object_name()
    task_name2 = random_object_name()
    try:
        task1 = tasks[task_name1]
        task1.create_or_alter(
            Task(
                name=task_name1,
                definition="select current_version()",
                schedule=timedelta(minutes=100),
                warehouse=db_parameters["warehouse"],
                allow_overlapping_execution=True,
                suspend_task_after_num_failures=100,
                config={"a": 1},
                # TODO: testing with error_integration requires setup of notification channel first.
                # error_integration=,
            )
        )
        task1_data = task1.fetch()
        assert task1_data.name == task_name1.upper()
        assert task1_data.definition == "select current_version()"
        assert task1_data.schedule == timedelta(minutes=100)
        assert task1_data.allow_overlapping_execution is True
        assert task1_data.suspend_task_after_num_failures == 100
        assert task1_data.config == {"a": 1}
        task1_data.schedule = Cron("* * * * *", "America/Los_Angeles")
        task1_data.warehouse = None
        task1_data.allow_overlapping_execution = False
        task1_data.suspend_task_after_num_failures = 200
        task1.create_or_alter(task1_data)
        task1_data_again = tasks[task_name1].fetch()
        assert task1_data_again.schedule == Cron("* * * * *", "America/Los_Angeles")
        assert task1_data_again.allow_overlapping_execution is False
        assert task1_data_again.warehouse is None
        assert task1_data.suspend_task_after_num_failures == 200

        task2 = tasks[task_name2]
        task2.create_or_alter(
            Task(
                name=task_name2,
                definition="select current_version()",
                user_task_managed_initial_warehouse_size="XSMALL",
                target_completion_interval=timedelta(minutes=10),
                serverless_task_min_statement_size="XSMALL",
                serverless_task_max_statement_size="SMALL",
                user_task_timeout_ms=1000000,
                comment="abc",
                predecessors=[task1.name],
                condition="1=1",
                session_parameters={"SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 80000, "SNOWPARK_LAZY_ANALYSIS": False},
            )
        )
        task2_data = next((t for t in tasks.iter() if t.name == task_name2.upper()), None)
        assert task2_data.name == task_name2.upper()
        assert task2_data.definition == "select current_version()"
        assert task2_data.user_task_managed_initial_warehouse_size == "XSMALL"
        assert task2_data.target_completion_interval == timedelta(minutes=10)
        assert task2_data.serverless_task_min_statement_size == "XSMALL"
        assert task2_data.serverless_task_max_statement_size == "SMALL"
        assert task2_data.user_task_timeout_ms == 1000000
        assert task2_data.comment == "abc"
        assert task2_data.condition == "1=1"
        assert task2_data.session_parameters == {
            "SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 80000,
            "SNOWPARK_LAZY_ANALYSIS": False,
        }
        assert [parse_table_name(x)[-1] for x in task2_data.predecessors] == [normalize_name(task_name1)]
        task2_data.definition = "select 3"
        task2_data.comment = "def"
        task2_data.condition = "2=2"
        task2_data.session_parameters = {"SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 90000, "TIMEZONE": "America/New_York"}
        task2.create_or_alter(task2_data)
        task2_data_again = task2.fetch()
        assert task2_data_again.name == task_name2.upper()
        assert task2_data_again.definition == "select 3"
        assert task2_data_again.comment == "def"
        assert task2_data_again.condition == "2=2"
        assert task2_data_again.session_parameters == {
            "SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 90000,
            "TIMEZONE": "America/New_York",
        }
        assert [parse_table_name(x)[-1] for x in task2_data_again.predecessors] == [normalize_name(task_name1)]

    finally:
        with suppress(Exception):
            tasks[task_name1].drop()
        with suppress(Exception):
            tasks[task_name2].drop()


def test_create_or_alter_config_and_schedule(tasks, db_parameters):
    task_name1 = random_object_name()
    try:
        task1 = tasks[task_name1]
        task1.create_or_alter(
            Task(
                name=task_name1, definition="select current_version()", schedule=timedelta(minutes=100), config={"a": 1}
            )
        )

        task1.create_or_alter(Task(name=task_name1, definition="select current_version()"))
        fetched = task1.fetch()
        assert fetched.config == {}
        assert fetched.schedule is None

        task1.create_or_alter(
            Task(
                name=task_name1, definition="select current_version()", schedule=timedelta(minutes=101), config={"b": 1}
            )
        )
        fetched_again = task1.fetch()
        assert fetched_again.config == {"b": 1}
        assert fetched_again.schedule == timedelta(minutes=101)
        task1.create_or_alter(
            Task(
                name=task_name1,
                definition="select current_version()",
                schedule=Cron("0 0 10-20 * TUE,THU", "America/Los_Angeles"),
            )
        )
        fetched_3rd = task1.fetch()
        fetched_3rd.schedule = Cron("0 0 10-20 * TUE,THU", "America/Los_Angeles")
    finally:
        with suppress(Exception):
            tasks[task_name1].drop()


@pytest.mark.min_sf_ver("8.27.0")
def test_create_or_alter_finalizer(database, schema, tasks):
    task_name1 = random_object_name()
    task_name2 = random_object_name()
    task_name3 = random_object_name()
    try:
        task1 = tasks[task_name1]
        task1.create_or_alter(
            Task(name=task_name1, definition="select current_version()", schedule=timedelta(minutes=100))
        )

        task2 = tasks[task_name2]
        task2.create_or_alter(Task(name=task_name2, definition="select current_version()", finalize=task1.name))

        fetched = task2.fetch()
        fetched1 = task1.fetch()
        assert fetched.finalize == f"{fetched1.name}"

        task3 = tasks[task_name3]
        task3.create_or_alter(
            Task(
                name=task_name3, definition="select current_version()", schedule=timedelta(minutes=100), config={"a": 1}
            )
        )

        task2.create_or_alter(Task(name=task_name2, definition="select current_version()", finalize=task3.name))

        fetched_again = task2.fetch()
        fetched3 = task3.fetch()
        assert fetched_again.finalize == f"{fetched3.name}"
    finally:
        with suppress(Exception):
            tasks[task_name1].drop()
            tasks[task_name2].drop()
            tasks[task_name3].drop()


@pytest.mark.snowpark
@pytest.mark.usefixtures("anaconda_package_available")
def test_create_or_alter_definition_python(tasks, db_parameters):
    from snowflake.snowpark import Session

    def foo1(session: Session) -> None:
        session.sql("select 1").collect()

    def foo2(session: Session) -> None:
        session.sql("select 2").collect()

    test_stage = random_object_name()
    test_task = random_object_name()
    tasks._connection.execute_string(f"create or replace stage {test_stage}")
    try:
        task1 = Task(
            test_task,
            StoredProcedureCall(foo1, stage_location=test_stage, packages=["snowflake-snowpark-python"]),
            warehouse=db_parameters["warehouse"],
        )
        task_ref = tasks[test_task]
        task_ref.create_or_alter(task1)
        try:
            fetched = task_ref.fetch()
            assert "foo1" in fetched.definition
            fetched.definition = StoredProcedureCall(
                foo2, stage_location=test_stage, packages=["snowflake-snowpark-python"]
            )
            task_ref.create_or_alter(fetched)
            fetched_again = task_ref.fetch()
            assert "foo2" in fetched_again.definition
        finally:
            task_ref.drop()
    finally:
        tasks._connection.execute_string(f"drop stage {test_stage}")



import pytest as pytest

from tests.utils import random_string

from .utils import create_service_function


TRANSLATE_DATA_TYPE_TO_BIND = {"INT": "FIXED"}

pytestmark = [pytest.mark.skip_gov]


def test_create_service_function_argument(temp_service_for_function, functions):
    for t in [["REAL"], ["INT"], ["BOOLEAN"], ["TIME"], ["REAL", "INT"]]:
        function_name = random_string(5, "test_func_")
        function_name_with_args = f"{function_name}({','.join(t)})"
        endpoint = "end-point-2" if t == "INT" else "ep1"
        try:
            f = create_service_function(function_name, t, t[0], endpoint, temp_service_for_function.name, functions)

            assert (
                f.fetch().to_dict().items()
                >= {
                    "function_type": "service-function",
                    "name": function_name.upper(),
                    "arguments": [
                        {"name": f"V_{i}", "datatype": TRANSLATE_DATA_TYPE_TO_BIND.get(v, v)} for i, v in enumerate(t)
                    ],
                    "returns": TRANSLATE_DATA_TYPE_TO_BIND.get(t[0], t[0]),
                    "max_batch_rows": 5,
                    "body": "/path/to/myapp",
                    "endpoint": endpoint,
                    "path": "/path/to/myapp",
                }.items()
            )
        finally:
            functions[function_name_with_args].drop()



import typing

from ._generated.api.sequence_api_base import SequenceCollectionBase, SequenceResourceBase


if typing.TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class SequenceCollection(SequenceCollectionBase):
    """Represents the collection operations on the Snowflake sequence resource.

    With this collection, you can create, iterate through, and search for sequences that you have access to
    in the current context.
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, SequenceResource)


class SequenceResource(SequenceResourceBase):
    """Represents a reference to a Snowflake sequence.

    With this sequence reference, you can fetch information about a sequence, as well as
    perform certain actions on it.
    """

    _plural_name = "sequences"



import copy

from tests.integ.utils import random_string

from .conftest import test_tag_template


def test_create_or_alter_tag(tags):
    tag_name = random_string(10, "test_tag_create_or_alter_")
    tag_def = copy.deepcopy(test_tag_template)
    tag_def.name = tag_name
    tag_def.comment = "created by test_create_or_alter_tag"

    tag = tags[tag_name]
    tag.create_or_alter(tag_def)
    try:
        fetched_tag = tag.fetch()
        assert fetched_tag.name.upper() == tag_name.upper()
        assert fetched_tag.allowed_values == tag_def.allowed_values
        assert fetched_tag.comment == tag_def.comment

        fetched_tag.comment = "altered by test_create_or_alter_tag"
        fetched_tag.allowed_values = ["new_value1", "new_value2", "new_value3"]
        tag.create_or_alter(fetched_tag)

        altered_tag = tag.fetch()
        assert altered_tag.name.upper() == tag_name.upper()
        assert altered_tag.allowed_values == ["new_value1", "new_value2", "new_value3"]
        assert altered_tag.comment == "altered by test_create_or_alter_tag"
    finally:
        tag.drop(if_exists=True)


def test_create_or_alter_unset_comment(tags):
    tag_name = random_string(10, "test_tag_unset_comment_")
    tag_def = copy.deepcopy(test_tag_template)
    tag_def.name = tag_name
    tag_def.comment = "comment to be removed"

    tag = tags.create(tag_def)
    try:
        updated_tag_def = copy.deepcopy(test_tag_template)
        updated_tag_def.name = tag_name
        updated_tag_def.comment = None
        tag.create_or_alter(updated_tag_def)

        fetched_tag = tag.fetch()
        assert fetched_tag.name.upper() == tag_name.upper()
        assert fetched_tag.comment is None
    finally:
        tag.drop(if_exists=True)



"""
Snowflake Stream API.

The Snowflake Stream API is a REST API that you can use to access, update, and perform certain actions on Stream resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import ConfigDict

from snowflake.core.stream._generated.models.stream_source import StreamSource


class StreamSourceStage(StreamSource):
    """A model object representing the StreamSourceStage resource.

    Constructs an object of type StreamSourceStage with the provided properties.

    Parameters
    __________

    name : str
        Name of the source whose changes are tracked by the stream

    database_name : str, optional
        Database name to which stream source type belongs. If not provided, database  name provided in the path param will be used.

    schema_name : str, optional
        Schema name to which stream source type belongs. If not provided, schema name provided in the path param will be used.
    """

    __properties = ["src_type", "name", "database_name", "schema_name"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> StreamSourceStage:
        """Create an instance of StreamSourceStage from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        _dict["src_type"] = StreamSource.get_child_model_discriminator_value("StreamSourceStage")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> StreamSourceStage:
        """Create an instance of StreamSourceStage from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return StreamSourceStage.model_validate(obj)

        _obj = StreamSourceStage.model_validate(
            {
                "name": obj.get("name"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
            }
        )

        return _obj


class StreamSourceStageModel(StreamSource):
    def __init__(
        self,
        name: str,  # optional properties
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
    ):
        """A model object representing the StreamSourceStage resource.

        Constructs an object of type StreamSourceStage with the provided properties.

        Parameters
        __________

        name : str
            Name of the source whose changes are tracked by the stream

        database_name : str, optional
            Database name to which stream source type belongs. If not provided, database  name provided in the path param will be used.

        schema_name : str, optional
            Schema name to which stream source type belongs. If not provided, schema name provided in the path param will be used.
        """
        super().__init__(
            name=name,
            database_name=database_name,
            schema_name=schema_name,
        )

    __properties = ["src_type", "name", "database_name", "schema_name"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return StreamSourceStage(
            name=self.name,
            database_name=self.database_name,
            schema_name=self.schema_name,
        )

    @classmethod
    def _from_model(cls, model) -> StreamSourceStageModel:
        return StreamSourceStageModel(
            name=model.name,
            database_name=model.database_name,
            schema_name=model.schema_name,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a StreamSourceStage.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the StreamSourceStage object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StreamSourceStageModel:
        """Create an instance of StreamSourceStage from a dict.

        This method constructs a StreamSourceStage object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        StreamSourceStage
            A StreamSourceStage object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(StreamSourceStage.from_dict(obj))


StreamSourceStage._model_class = StreamSourceStageModel



import os
import tempfile

from io import BytesIO
from pathlib import Path
from textwrap import dedent
from typing import Iterator

import pytest

from snowflake.core.stage import Stage, StageResource
from tests.integ.utils import random_string


@pytest.fixture(scope="module")
def streamlit_main_file():
    return "app.py"


@pytest.fixture(scope="module")
def streamlit_stage_with_file(session, stages, streamlit_main_file) -> Iterator[StageResource]:
    stage_name = random_string(8, "test_streamlit_stage_")
    stage = Stage(name=stage_name, kind="PERMANENT")

    st = stages.create(stage)
    try:
        file_path = f"@{st.name}/{streamlit_main_file}"

        session.file.put_stream(
            BytesIO(
                dedent(
                    """
                    import streamlit as st
                    st.write("Hello from streamlit test")
                    """
                ).encode()
            ),
            file_path,
            auto_compress=False,
        )

        yield st
    finally:
        st.drop()


@pytest.fixture(scope="module")
def git_repository():
    """Create a temporary Git repository with sample streamlit files.

    If TEST_GIT_REPO_URL environment variable is set, uses that repository.
    Otherwise, attempts to create a GitHub repository using GitHub CLI or API.
    If neither is available, returns a local repository (tests will be skipped).
    """
    import shutil
    import subprocess

    temp_dir = tempfile.mkdtemp(prefix="streamlit_test_git_")
    repo_path = Path(temp_dir)
    remote_repo_url = None
    repo_name = None
    github_org = None

    try:
        try:
            subprocess.run(["git", "--version"], capture_output=True, check=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            yield {
                "path": str(repo_path),
                "commit_hash": None,
                "main_file": "app.py",
                "files": [],
                "remote_url": None,
                "repo_name": None,
                "github_org": None,
            }
            return

        subprocess.run(["git", "init"], cwd=repo_path, check=True)

        app_py = repo_path / "app.py"
        app_py.write_text(
            dedent("""
            import streamlit as st

            st.title("Test Streamlit App")
            st.write("Hello from Git repository!")

            # Add some interactivity
            name = st.text_input("Enter your name:")
            if name:
                st.write(f"Hello, {name}!")
        """)
        )

        utils_py = repo_path / "utils.py"
        utils_py.write_text(
            dedent("""
            def get_greeting(name):
                return f"Hello from utils, {name}!"
        """)
        )

        config_yaml = repo_path / "config.yaml"
        config_yaml.write_text(
            dedent("""
            app:
              title: "Test Streamlit App"
              description: "A test app from Git repository"
        """)
        )

        gitignore = repo_path / ".gitignore"
        gitignore.write_text("__pycache__/\n*.pyc\n.pytest_cache/\n")

        git_env = {
            **os.environ,
            "GIT_AUTHOR_NAME": "Test User",
            "GIT_AUTHOR_EMAIL": "test@example.com",
            "GIT_COMMITTER_NAME": "Test User",
            "GIT_COMMITTER_EMAIL": "test@example.com",
        }

        subprocess.run(["git", "add", "."], cwd=repo_path, check=True)
        subprocess.run(
            ["git", "commit", "-m", "Initial commit with streamlit files"],
            cwd=repo_path,
            check=True,
            env=git_env,
        )

        readme = repo_path / "README.md"
        readme.write_text("# Test Streamlit Repository\n\nThis is a test repository for streamlit integration tests.")

        subprocess.run(["git", "add", "README.md"], cwd=repo_path, check=True)
        subprocess.run(
            ["git", "commit", "-m", "Add README"],
            cwd=repo_path,
            check=True,
            env=git_env,
        )

        result = subprocess.run(["git", "rev-parse", "HEAD"], cwd=repo_path, capture_output=True, text=True, check=True)
        commit_hash = result.stdout.strip()

        subprocess.run(["git", "branch", "-M", "main"], cwd=repo_path, check=True)

        test_repo_url = os.getenv("TEST_GIT_REPO_URL")
        if test_repo_url:
            remote_repo_url = test_repo_url
            try:
                subprocess.run(["git", "remote", "add", "origin", remote_repo_url], cwd=repo_path, check=True)
                subprocess.run(["git", "push", "-u", "origin", "main"], cwd=repo_path, check=True, capture_output=True)
            except subprocess.CalledProcessError:
                pass
        else:
            try:
                github_org = os.getenv("GITHUB_ORG", "snowflakedb")
                repo_name = f"streamlit-test-{random_string(8, '')}"
                result = subprocess.run(
                    ["gh", "repo", "create", f"{github_org}/{repo_name}", "--public", "--source", ".", "--push"],
                    cwd=repo_path,
                    check=True,
                    capture_output=True,
                    text=True,
                )
                remote_repo_url = f"https://github.com/{github_org}/{repo_name}.git"
            except (subprocess.CalledProcessError, FileNotFoundError):
                github_token = os.getenv("GITHUB_TOKEN") or os.getenv("GH_TOKEN")
                if github_token:
                    try:
                        import requests

                        github_org = os.getenv("GITHUB_ORG", "snowflakedb")
                        repo_name = f"streamlit-test-{random_string(8, '')}"
                        headers = {
                            "Authorization": f"token {github_token}",
                            "Accept": "application/vnd.github.v3+json",
                        }
                        data = {
                            "name": repo_name,
                            "description": "Temporary test repository for Streamlit integration tests",
                            "private": False,
                            "auto_init": False,
                        }
                        response = requests.post(
                            f"https://api.github.com/orgs/{github_org}/repos",
                            headers=headers,
                            json=data,
                            timeout=10,
                        )
                        response.raise_for_status()
                        remote_repo_url = response.json()["clone_url"]

                        subprocess.run(
                            ["git", "remote", "add", "origin", remote_repo_url],
                            cwd=repo_path,
                            check=True,
                        )
                        auth_url = remote_repo_url.replace("https://", f"https://{github_token}@")
                        subprocess.run(
                            ["git", "remote", "set-url", "origin", auth_url],
                            cwd=repo_path,
                            check=True,
                        )
                        subprocess.run(
                            ["git", "push", "-u", "origin", "main"],
                            cwd=repo_path,
                            check=True,
                            capture_output=True,
                        )
                    except Exception:
                        pass

        yield {
            "path": str(repo_path),
            "commit_hash": commit_hash,
            "main_file": "app.py",
            "files": ["app.py", "utils.py", "config.yaml", "README.md", ".gitignore"],
            "remote_url": remote_repo_url,
            "repo_name": repo_name,
            "github_org": github_org,
        }

    finally:
        if repo_name and github_org:
            github_token = os.getenv("GITHUB_TOKEN") or os.getenv("GH_TOKEN")
            if github_token:
                try:
                    import requests

                    headers = {
                        "Authorization": f"token {github_token}",
                        "Accept": "application/vnd.github.v3+json",
                    }
                    requests.delete(
                        f"https://api.github.com/repos/{github_org}/{repo_name}",
                        headers=headers,
                        timeout=10,
                    )
                except Exception:
                    pass
            else:
                try:
                    subprocess.run(
                        ["gh", "repo", "delete", f"{github_org}/{repo_name}", "--yes"],
                        check=True,
                        capture_output=True,
                    )
                except Exception:
                    pass

        shutil.rmtree(temp_dir, ignore_errors=True)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import pytest

from snowflake.core.cortex.search_service import QueryRequest


TEST_SERVICE_NAME = "SNOWPY_TEST_SERVICE"

API_NOT_ENABLED = "Cortex Search API is not enabled"

EXACTLY_ONE_OF = "The request must contain exactly one of 'query' or 'multi_index_query'"
pytestmark = [pytest.mark.skip_gov]


@pytest.fixture(autouse=True)
def setup_cortex_search_service(connection, database, schema, warehouse, backup_warehouse_fixture):
    del backup_warehouse_fixture
    _database_name = database.name
    _schema_name = schema.name
    with connection.cursor() as cursor:
        cursor.execute(f"use warehouse {warehouse.name}").fetchone()

        test_table_name = f"{_database_name}.{_schema_name}.SNOWPY_TEST_TABLE"
        # Base Table
        cursor.execute(
            f"CREATE OR REPLACE TABLE {test_table_name} (id NUMBER AUTOINCREMENT, col1 VARCHAR, col2 VARCHAR)"
        )

        rows = ",".join(["('hi', 'hello')"] * 20)
        cursor.execute(f"INSERT INTO {test_table_name} (col1, col2) VALUES {rows}")

        # Cortex Search Service
        cursor.execute(
            f"CREATE OR REPLACE CORTEX SEARCH SERVICE {_database_name}.{_schema_name}.{TEST_SERVICE_NAME} "
            f"ON col1 TARGET_LAG='1 minute' WAREHOUSE={warehouse.name} "
            f"AS SELECT id, col1, col2 FROM {test_table_name}"
        )

        try:
            yield
        finally:
            cursor.execute(f"DROP CORTEX SEARCH SERVICE {_database_name}.{_schema_name}.{TEST_SERVICE_NAME}")
            cursor.execute(f"DROP TABLE {test_table_name}")


@pytest.fixture(autouse=True)
def precheck_cortex_search_enabled(cortex_search_services, setup_cortex_search_service):
    del setup_cortex_search_service
    try:
        cortex_search_services[TEST_SERVICE_NAME].search("hi", ["col1", "col2"], limit=5)
    except Exception as err:
        if API_NOT_ENABLED in err.body:
            pytest.xfail(API_NOT_ENABLED)
        raise


def test_search(cortex_search_services):
    resp = cortex_search_services[TEST_SERVICE_NAME].search("hi", ["col1", "col2"], limit=5)
    assert len(resp.results) == 5
    for row in resp.results:
        assert row["col1"] is not None
        assert row["col2"] is not None


def test_search_optionalized(cortex_search_services):
    # because no column is specified, we default to displaying the indexed search column
    resp = cortex_search_services[TEST_SERVICE_NAME].search(query="hi", limit=5)
    assert len(resp.results) == 5
    for row in resp.results:
        assert row["COL1"] is not None

    # If no query input is specified (query or multi_index_query) as well as no columns, then we should see an error.
    try:
        cortex_search_services[TEST_SERVICE_NAME].search(limit=5)
    except Exception as err:
        if EXACTLY_ONE_OF not in err.body:
            pytest.xfail(EXACTLY_ONE_OF)


def test_search_collection(cortex_search_services):
    resp = cortex_search_services.search(
        TEST_SERVICE_NAME, QueryRequest.from_dict({"query": "hi", "columns": ["col1", "col2"], "limit": 5})
    )

    assert len(resp.results) == 5
    for row in resp.results:
        assert row["col1"] is not None
        assert row["col2"] is not None


def test_experimental_arg(cortex_search_services) -> None:
    resp = cortex_search_services[TEST_SERVICE_NAME].search(
        "hi", ["col1", "col2"], limit=5, experimental={"debug": True}
    )
    assert len(resp.results) == 5
    print(resp)
    for row in resp.results:
        assert row["col1"] is not None
        assert row["col2"] is not None
        assert "@DEBUG_PER_RESULT" in row
        assert "TopicalityScore" in row["@DEBUG_PER_RESULT"]



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictStr

from snowflake.core.cortex.analyst_service._generated.models.sql_inputs import SqlInputs, SqlInputsModel
from snowflake.core.cortex.analyst_service._generated.models.tableau_input import TableauInput, TableauInputModel
from snowflake.core.cortex.analyst_service._generated.models.tds_input import TdsInput, TdsInputModel


class PreSelectionRequest(BaseModel):
    """A model object representing the PreSelectionRequest resource.

    Constructs an object of type PreSelectionRequest with the provided properties.

    Parameters
    __________
    tds_source : list[TdsInput]

    warehouse : str
        Warehouse name to use for processing.
    sql_source : SqlInputs, optional

    tableau_input : TableauInput, optional

    experimental : str, optional
        JSON serialized string of experimental API fields (undocumented).
    """

    tds_source: List[TdsInput]

    sql_source: Optional[SqlInputs] = None

    warehouse: StrictStr

    tableau_input: Optional[TableauInput] = None

    experimental: Optional[StrictStr] = None

    __properties = ["tds_source", "sql_source", "warehouse", "tableau_input", "experimental"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PreSelectionRequest:
        """Create an instance of PreSelectionRequest from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in tds_source (list)
        _items = []
        if self.tds_source:
            for _item in self.tds_source:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tds_source"] = _items

        # override the default output from pydantic by calling `to_dict()` of sql_source
        if self.sql_source:
            _dict["sql_source"] = self.sql_source.to_dict()

        # override the default output from pydantic by calling `to_dict()` of tableau_input
        if self.tableau_input:
            _dict["tableau_input"] = self.tableau_input.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> PreSelectionRequest:
        """Create an instance of PreSelectionRequest from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return PreSelectionRequest.model_validate(obj)

        _obj = PreSelectionRequest.model_validate(
            {
                "tds_source": [TdsInput.from_dict(_item) for _item in obj.get("tds_source")]
                if obj.get("tds_source") is not None
                else None,
                "sql_source": SqlInputs.from_dict(obj.get("sql_source")) if obj.get("sql_source") is not None else None,
                "warehouse": obj.get("warehouse"),
                "tableau_input": TableauInput.from_dict(obj.get("tableau_input"))
                if obj.get("tableau_input") is not None
                else None,
                "experimental": obj.get("experimental"),
            }
        )

        return _obj


class PreSelectionRequestModel:
    def __init__(
        self,
        tds_source: list[TdsInput],
        warehouse: str,
        # optional properties
        sql_source: Optional[SqlInputs] = None,
        tableau_input: Optional[TableauInput] = None,
        experimental: Optional[str] = None,
    ):
        """A model object representing the PreSelectionRequest resource.

        Constructs an object of type PreSelectionRequest with the provided properties.

        Parameters
        __________
        tds_source : list[TdsInput]

        warehouse : str
            Warehouse name to use for processing.
        sql_source : SqlInputs, optional

        tableau_input : TableauInput, optional

        experimental : str, optional
            JSON serialized string of experimental API fields (undocumented).
        """
        self.tds_source = tds_source
        self.sql_source = sql_source
        self.warehouse = warehouse
        self.tableau_input = tableau_input
        self.experimental = experimental

    __properties = ["tds_source", "sql_source", "warehouse", "tableau_input", "experimental"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return PreSelectionRequest(
            tds_source=[x._to_model() for x in self.tds_source] if self.tds_source is not None else None,
            sql_source=self.sql_source._to_model() if self.sql_source is not None else None,
            warehouse=self.warehouse,
            tableau_input=self.tableau_input._to_model() if self.tableau_input is not None else None,
            experimental=self.experimental,
        )

    @classmethod
    def _from_model(cls, model) -> PreSelectionRequestModel:
        return PreSelectionRequestModel(
            tds_source=[TdsInputModel._from_model(x) for x in model.tds_source] if model.tds_source else None,
            sql_source=SqlInputsModel._from_model(model.sql_source) if model.sql_source else None,
            warehouse=model.warehouse,
            tableau_input=TableauInputModel._from_model(model.tableau_input) if model.tableau_input else None,
            experimental=model.experimental,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a PreSelectionRequest.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the PreSelectionRequest object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> PreSelectionRequestModel:
        """Create an instance of PreSelectionRequest from a dict.

        This method constructs a PreSelectionRequest object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        PreSelectionRequest
            A PreSelectionRequest object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(PreSelectionRequest.from_dict(obj))


PreSelectionRequest._model_class = PreSelectionRequestModel



"""
Snowflake Stream API.

The Snowflake Stream API is a REST API that you can use to access, update, and perform certain actions on Stream resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictStr, field_validator
from typing_extensions import Annotated

from snowflake.core.stream._generated.models.stream_source import StreamSource, StreamSourceModel


class Stream(BaseModel):
    """A model object representing the Stream resource.

    Constructs an object of type Stream with the provided properties.

    Parameters
    __________
    name : str
        Name of the stream
    stream_source : StreamSource

    created_on : datetime, optional
        Date and time when the stream was created — **Read-only:** *any user-provided value will be ignored.*
    comment : str, optional
        user comment associated to an object in the dictionary
    database_name : str, optional
        Database in which the stream is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the stream is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the stream — **Read-only:** *any user-provided value will be ignored.*
    table_name : str, optional
        Table name whose changes are tracked by the stream — **Read-only:** *any user-provided value will be ignored.*
    stale : bool, optional
        Specifies whether the stream is stale or not — **Read-only:** *any user-provided value will be ignored.*
    mode : str, optional
        Mode of the stream. Possible values include: APPEND_ONLY, INSERT_ONLY. For streams on tables, the column displays DEFAULT — **Read-only:** *any user-provided value will be ignored.*
    stale_after : datetime, optional
        Timestamp when the stream became stale or may become stale if not consumed — **Read-only:** *any user-provided value will be ignored.*
    invalid_reason : str, optional
        Reason why the stream cannot be queried successfully. This column supports future functionality. Currently, the only value returned is N/A — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the stream — **Read-only:** *any user-provided value will be ignored.*
    type : str, optional
        Type of the stream; currently DELTA only — **Read-only:** *any user-provided value will be ignored.*
    """

    created_on: Optional[datetime] = None

    name: Annotated[str, Field(strict=True)]

    stream_source: StreamSource

    comment: Optional[StrictStr] = None

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    owner: Optional[Annotated[str, Field(strict=True)]] = None

    table_name: Optional[Annotated[str, Field(strict=True)]] = None

    stale: Optional[StrictBool] = None

    mode: Optional[StrictStr] = None

    stale_after: Optional[datetime] = None

    invalid_reason: Optional[StrictStr] = None

    owner_role_type: Optional[Annotated[str, Field(strict=True)]] = None

    type: Optional[StrictStr] = None

    __properties = [
        "created_on",
        "name",
        "stream_source",
        "comment",
        "database_name",
        "schema_name",
        "owner",
        "table_name",
        "stale",
        "mode",
        "stale_after",
        "invalid_reason",
        "owner_role_type",
        "type",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner")
    def owner_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("table_name")
    def table_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner_role_type")
    def owner_role_type_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Stream:
        """Create an instance of Stream from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "table_name",
                    "stale",
                    "mode",
                    "stale_after",
                    "invalid_reason",
                    "owner_role_type",
                    "type",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of stream_source
        if self.stream_source:
            _dict["stream_source"] = self.stream_source.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Stream:
        """Create an instance of Stream from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Stream.model_validate(obj)

        _obj = Stream.model_validate(
            {
                "created_on": obj.get("created_on"),
                "name": obj.get("name"),
                "stream_source": StreamSource.from_dict(obj.get("stream_source"))
                if obj.get("stream_source") is not None
                else None,
                "comment": obj.get("comment"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "table_name": obj.get("table_name"),
                "stale": obj.get("stale"),
                "mode": obj.get("mode"),
                "stale_after": obj.get("stale_after"),
                "invalid_reason": obj.get("invalid_reason"),
                "owner_role_type": obj.get("owner_role_type"),
                "type": obj.get("type"),
            }
        )

        return _obj


class StreamModel:
    def __init__(
        self,
        name: str,
        stream_source: StreamSource,
        # optional properties
        created_on: Optional[datetime] = None,
        comment: Optional[str] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        table_name: Optional[str] = None,
        stale: Optional[bool] = None,
        mode: Optional[str] = None,
        stale_after: Optional[datetime] = None,
        invalid_reason: Optional[str] = None,
        owner_role_type: Optional[str] = None,
        type: Optional[str] = None,
    ):
        """A model object representing the Stream resource.

        Constructs an object of type Stream with the provided properties.

        Parameters
        __________
        name : str
            Name of the stream
        stream_source : StreamSource

        created_on : datetime, optional
            Date and time when the stream was created.
        comment : str, optional
            user comment associated to an object in the dictionary
        database_name : str, optional
            Database in which the stream is stored
        schema_name : str, optional
            Schema in which the stream is stored
        owner : str, optional
            Role that owns the stream
        table_name : str, optional
            Table name whose changes are tracked by the stream
        stale : bool, optional
            Specifies whether the stream is stale or not
        mode : str, optional
            Mode of the stream. Possible values include: APPEND_ONLY, INSERT_ONLY. For streams on tables, the column displays DEFAULT.
        stale_after : datetime, optional
            Timestamp when the stream became stale or may become stale if not consumed.
        invalid_reason : str, optional
            Reason why the stream cannot be queried successfully. This column supports future functionality. Currently, the only value returned is N/A.
        owner_role_type : str, optional
            The type of role that owns the stream
        type : str, optional
            Type of the stream; currently DELTA only.
        """
        self.created_on = created_on
        self.name = name
        self.stream_source = stream_source
        self.comment = comment
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.table_name = table_name
        self.stale = stale
        self.mode = mode
        self.stale_after = stale_after
        self.invalid_reason = invalid_reason
        self.owner_role_type = owner_role_type
        self.type = type

    __properties = [
        "created_on",
        "name",
        "stream_source",
        "comment",
        "database_name",
        "schema_name",
        "owner",
        "table_name",
        "stale",
        "mode",
        "stale_after",
        "invalid_reason",
        "owner_role_type",
        "type",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Stream(
            created_on=self.created_on,
            name=self.name,
            stream_source=self.stream_source._to_model() if self.stream_source is not None else None,
            comment=self.comment,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            table_name=self.table_name,
            stale=self.stale,
            mode=self.mode,
            stale_after=self.stale_after,
            invalid_reason=self.invalid_reason,
            owner_role_type=self.owner_role_type,
            type=self.type,
        )

    @classmethod
    def _from_model(cls, model) -> StreamModel:
        return StreamModel(
            created_on=model.created_on,
            name=model.name,
            stream_source=StreamSourceModel._from_model(model.stream_source) if model.stream_source else None,
            comment=model.comment,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            table_name=model.table_name,
            stale=model.stale,
            mode=model.mode,
            stale_after=model.stale_after,
            invalid_reason=model.invalid_reason,
            owner_role_type=model.owner_role_type,
            type=model.type,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Stream.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Stream object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> StreamModel:
        """Create an instance of Stream from a dict.

        This method constructs a Stream object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Stream
            A Stream object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Stream.from_dict(obj))


Stream._model_class = StreamModel



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterator

import pytest

from snowflake.core._common import CreateMode
from snowflake.core.exceptions import APIError, ConflictError
from snowflake.core.procedure import (
    Argument,
    ColumnType,
    JavaFunction,
    JavaScriptFunction,
    Procedure,
    PythonFunction,
    ReturnDataType,
    ReturnTable,
    ScalaFunction,
    SQLFunction,
)
from snowflake.core.procedure._generated import FunctionLanguage
from tests.utils import random_string


@pytest.fixture
def temp_dir() -> Iterator[Path]:
    with TemporaryDirectory() as temp_dir:
        yield Path(temp_dir)


@pytest.fixture
def staged_handler(temp_dir, temp_stage) -> Iterator[tuple[str, str]]:
    handler_path = temp_dir / "handler.py"
    with open(handler_path, "w") as handler:
        handler.write("""
def run():
    return 'OK'
        """)
    temp_stage.put(handler_path, "/")
    stage_fqn = f"{temp_stage.database.name.upper()}.{temp_stage.schema.name.upper()}.{temp_stage.name.upper()}"
    yield (
        f"@{stage_fqn}/handler.py",
        "handler.run",
    )


def test_create_javascript_proc(procedures):
    javascript_proc_no_args = Procedure(
        name="javascript_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="FLOAT", nullable=False),
        language_config=JavaScriptFunction(),
        body="""return 3.14;""",
    )
    try:
        proc = procedures.create(javascript_proc_no_args, mode=CreateMode.or_replace, copy_grants=True)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVASCRIPT_PROC_NO_ARGS"
        assert proc_fetch_result.language_config == JavaScriptFunction(called_on_null_input=True)
        assert proc_fetch_result.return_type == ReturnDataType(datatype="FLOAT", nullable=False)
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    # JavaScript Procedure - Arguments
    javascript_proc_with_args = Procedure(
        name="javascript_proc_with_args",
        arguments=[Argument(name="TSV", datatype="VARCHAR", default="this will not work without input")],
        return_type=ReturnDataType(datatype="TIMESTAMP_LTZ"),
        language_config=JavaScriptFunction(),
        body="""// Convert the input varchar to a TIMESTAMP_LTZ.
                var sql_command = "SELECT '" + TSV + "'::TIMESTAMP_LTZ;";
                var stmt = snowflake.createStatement( {sqlText: sql_command} );
                var resultSet = stmt.execute();
                resultSet.next();
                // Retrieve the TIMESTAMP_LTZ and store it in an SfDate variable.
                var my_sfDate = resultSet.getColumnValue(1);

                f = 3.1415926;

                // Specify that we'd like position-based binding.
                sql_command = `INSERT INTO table1 VALUES(:1, :2, :3, :4, :5, :6, :7, :8);`
                // Bind a VARCHAR, a TIMESTAMP_LTZ, a numeric to our INSERT statement.
                result = snowflake.execute(
                    {
                    sqlText: sql_command,
                    binds: [TSV, my_sfDate, f, f, f, my_sfDate, my_sfDate, '12:30:00.123' ]
                    }
                    );

                return my_sfDate;
                """,
    )
    try:
        proc = procedures.create(javascript_proc_with_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVASCRIPT_PROC_WITH_ARGS"
    finally:
        proc.drop()

    javascript_proc_strict = Procedure(
        name="javascript_proc_strict",
        arguments=[Argument(name="FLOAT_PARAM1", datatype="FLOAT", default=2.2)],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=JavaScriptFunction(),
        is_secure=True,
        body="""var sql_command =
                 "INSERT INTO stproc_test_table1 (num_col1) VALUES (" + FLOAT_PARAM1 + ")";
                try {
                    snowflake.execute (
                        {sqlText: sql_command}
                    );
                    return "Succeeded.";   // Return a success/error indicator.
                }
                catch (err)  {
                    return "Failed: " + err;   // Return a success/error indicator.
                }
            """,
    )
    try:
        proc = procedures.create(javascript_proc_strict, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVASCRIPT_PROC_STRICT"
        assert proc_fetch_result.arguments == [Argument(name="FLOAT_PARAM1", datatype="FLOAT", default=2.2)]
        assert proc_fetch_result.return_type == ReturnDataType(datatype="VARCHAR", nullable=True)
        assert proc_fetch_result.language_config == JavaScriptFunction(called_on_null_input=True)
        assert proc_fetch_result.is_secure is True
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    javascript_proc_execute_as_and_comment = Procedure(
        name="javascript_proc_execute_as_and_comment",
        comment="comment",
        execute_as="OWNER",
        arguments=[Argument(name="FLOAT_PARAM1", datatype="FLOAT")],
        return_type=ReturnDataType(datatype="VARCHAR", nullable=False),
        language_config=JavaScriptFunction(),
        is_secure=True,
        body="""  var row_count = 0;
                  // Dynamically compose the SQL statement to execute.
                  var sql_command = "select count(*) from " + TABLE_NAME;
                  // Run the statement.
                  var stmt = snowflake.createStatement(
                         {
                         sqlText: sql_command
                         }
                      );
                  var res = stmt.execute();
                  // Get back the row count. Specifically, ...
                  // ... get the first (and in this case only) row from the result set ...
                  res.next();
                  // ... and then get the returned value, which in this case is the number of
                  // rows in the table.
                  row_count = res.getColumnValue(1);
                  return row_count;
            """,
    )
    try:
        proc = procedures.create(javascript_proc_execute_as_and_comment, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVASCRIPT_PROC_EXECUTE_AS_AND_COMMENT"
        assert proc_fetch_result.arguments == [Argument(name="FLOAT_PARAM1", datatype="FLOAT")]
        assert proc_fetch_result.return_type == ReturnDataType(datatype="VARCHAR", nullable=False)
        assert proc_fetch_result.language_config == JavaScriptFunction(called_on_null_input=True)
        assert proc_fetch_result.execute_as == "OWNER"
        assert proc_fetch_result.comment == "comment"
        assert proc_fetch_result.is_secure is True
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()


def test_create_sql_proc(procedures):
    sql_proc_no_args = Procedure(
        name="sql_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="NUMBER", nullable=False),
        language_config=SQLFunction(),
        body="""BEGIN return 3; END""",
    )
    try:
        proc = procedures.create(sql_proc_no_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "SQL_PROC_NO_ARGS"
        assert proc_fetch_result.return_type == ReturnDataType(datatype="NUMBER", nullable=False)
        assert proc_fetch_result.language_config == SQLFunction(called_on_null_input=True)
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    # Sql Procedure - Arguments
    sql_proc_with_args = Procedure(
        name="sql_proc_with_args",
        arguments=[
            Argument(name="number1", datatype="NUMBER", default=1),
            Argument(name="number2", datatype="NUMBER", default=2),
        ],
        return_type=ReturnDataType(datatype="BOOLEAN", nullable=False),
        language_config=SQLFunction(),
        body="""BEGIN
                    IF (number1 > number2) THEN
                        RETURN true;
                    ELSE
                        RETURN false;
                    END IF;
                END
            """,
    )
    try:
        proc = procedures.create(sql_proc_with_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "SQL_PROC_WITH_ARGS"
    finally:
        proc.drop()

    # Sql Procedure - Table Function
    sql_proc_table_func = Procedure(
        name="sql_proc_table_func",
        arguments=[Argument(name="id", datatype="VARCHAR")],
        return_type=ReturnTable(
            column_list=[ColumnType(name="id", datatype="NUMBER"), ColumnType(name="price", datatype="NUMBER")]
        ),
        language_config=SQLFunction(),
        body="""DECLARE
                    res RESULTSET DEFAULT (SELECT * FROM invoices WHERE id = :id);
                BEGIN
                    RETURN TABLE(res);
                END;
            """,
    )
    try:
        proc = procedures.create(sql_proc_table_func, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "SQL_PROC_TABLE_FUNC"
    finally:
        proc.drop()

    # Sql Procedure
    sql_proc_parameters = Procedure(
        name="sql_proc_parameters",
        arguments=[Argument(name="id", datatype="VARCHAR")],
        is_secure=True,
        execute_as="CALLER",
        return_type=ReturnTable(
            column_list=[ColumnType(name="id", datatype="NUMBER"), ColumnType(name="price", datatype="NUMBER")]
        ),
        language_config=SQLFunction(called_on_null_input=True),
        body="""DECLARE
                    res RESULTSET DEFAULT (SELECT * FROM invoices WHERE id = :id);
                BEGIN
                    RETURN TABLE(res);
                END;
            """,
    )
    try:
        proc = procedures.create(sql_proc_parameters, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "SQL_PROC_PARAMETERS"
    finally:
        proc.drop()


def test_create_java_proc(procedures, maven_snowpark_jar_available):
    del maven_snowpark_jar_available
    # Java Procedure - No Arguments
    java_proc_no_args = Procedure(
        name="java_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=JavaFunction(
            runtime_version="11", handler="Test.testProc", packages=["com.snowflake:snowpark:latest"]
        ),
        body="""class Test {
                    public String testProc(com.snowflake.snowpark.Session session) {
                        return "ret string";
                    }
                }
            """,
    )
    try:
        proc = procedures.create(java_proc_no_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVA_PROC_NO_ARGS"
    finally:
        proc.drop()

    # Java Procedure - Table Function
    java_proc_table_func = Procedure(
        name="java_proc_table_func",
        arguments=[
            Argument(name="tableName", datatype="VARCHAR", default="this will not work without input"),
            Argument(name="role", datatype="VARCHAR"),
            Argument(name="int1", datatype="NUMBER", default=1),
            Argument(name="ts1", datatype="TIMESTAMP_LTZ"),
        ],
        return_type=ReturnTable(column_list=[ColumnType(name="Name", datatype="VARCHAR")]),
        language_config=JavaFunction(
            runtime_version="11", handler="FilterClass.filterByRole", packages=["com.snowflake:snowpark:latest"]
        ),
        body="""import com.snowflake.snowpark_java.*;
                public class FilterClass {
                    public DataFrame filterByRole(Session session,
                                                    String tableName, String role, int int1, String ts1) {
                        DataFrame table = session.table(tableName);
                        DataFrame filteredRows = table.filter(Functions.col("role").equal_to(Functions.lit(role)));
                        return filteredRows;
                    }
                }
            """,
    )
    try:
        proc = procedures.create(java_proc_table_func, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        javafun = proc_fetch_result.language_config
        assert proc_fetch_result.name == "JAVA_PROC_TABLE_FUNC"
        assert proc_fetch_result.arguments == [
            Argument(name="TABLENAME", datatype="VARCHAR", default="this will not work without input"),
            Argument(name="ROLE", datatype="VARCHAR"),
            Argument(name="INT1", datatype="NUMBER", default=1),
            Argument(name="TS1", datatype="TIMESTAMP_LTZ"),
        ]
        assert proc_fetch_result.return_type == ReturnTable(column_list=[ColumnType(name="NAME", datatype="VARCHAR")])
        assert javafun.runtime_version == "11"
        assert javafun.handler == "FilterClass.filterByRole"
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    # Java Procedure - Arguments
    java_proc_with_args = Procedure(
        name="java_proc_with_args",
        arguments=[
            Argument(name="bin", datatype="BINARY", default="534E4F57"),
            Argument(name="bool", datatype="BOOLEAN", default="FALSE"),
        ],
        return_type=ReturnDataType(name="ret", datatype="VARCHAR"),
        language_config=JavaFunction(
            runtime_version="11", handler="Test.retGeo", packages=["com.snowflake:snowpark:latest"]
        ),
        # we're returning string since the GEOGRAPHY datatype will accept strings
        body="""class Test {
                    public String retGeo(com.snowflake.snowpark.Session session, String binary, boolean bool) {
                        if(binary.equals("534E4F57")) {
                            return binary;
                        } else {
                            return String.valueOf(bool);
                        }
                    }
                }
            """,
    )
    try:
        proc = procedures.create(java_proc_with_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "JAVA_PROC_WITH_ARGS"
    finally:
        proc.drop()

    # Java Procedure
    java_proc_parameters = Procedure(
        name="java_proc_with_args",
        arguments=[
            Argument(name="geography", datatype="GEOGRAPHY"),
            Argument(name="int1", datatype="NUMBER", default="1"),
            Argument(name="arr1", datatype="ARRAY"),
        ],
        is_secure=False,
        comment="comment",
        execute_as="CALLER",
        return_type=ReturnDataType(name="Geography", datatype="GEOGRAPHY"),
        language_config=JavaFunction(
            called_on_null_input=False,
            runtime_version="11",
            handler="Test.retGeo",
            packages=["com.snowflake:snowpark:latest"],
        ),
        # we're returning string since the GEOGRAPHY datatype will accept strings
        body="""class Test {
                    public String retGeo(com.snowflake.snowpark.Session session,
                                            String geography, int int1, String[] arr1) {
                        return geography;
                    }
                }
        """,
    )
    try:
        proc = procedures.create(java_proc_parameters, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        javafn = proc_fetch_result.language_config
        assert proc_fetch_result.name == "JAVA_PROC_WITH_ARGS"
        assert proc_fetch_result.arguments == [
            Argument(name="GEOGRAPHY", datatype="GEOGRAPHY"),
            Argument(name="INT1", datatype="NUMBER", default="1"),
            Argument(name="ARR1", datatype="ARRAY"),
        ]
        assert proc_fetch_result.return_type == ReturnDataType(name="Geography", datatype="GEOGRAPHY", nullable=True)
        assert javafn.runtime_version == "11"
        assert javafn.handler == "Test.retGeo"
        assert proc_fetch_result.is_builtin is True
        assert proc_fetch_result.comment == "comment"
        assert proc_fetch_result.execute_as == "CALLER"
        assert proc_fetch_result.is_secure is False
    finally:
        proc.drop()


@pytest.mark.min_sf_ver("9.39.0")
def test_create_and_fetch_python_proc_staged_handler(procedures, staged_handler):
    proc_name = random_string(10, "python_proc_staged_handler_")
    file_path, handler = staged_handler
    language_config = PythonFunction(
        runtime_version="3.13", handler=handler, packages=["snowflake-snowpark-python"], imports=[file_path]
    )
    python_proc_staged_handler = Procedure(
        name=proc_name,
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=language_config,
        execute_as="OWNER",
    )
    try:
        proc = procedures.create(python_proc_staged_handler)
        fetched_proc = proc.fetch()
        assert fetched_proc.name == proc_name.upper()
        assert fetched_proc.arguments == []
        assert fetched_proc.return_type == ReturnDataType(datatype="VARCHAR", nullable=True)
        assert fetched_proc.language_config.runtime_version == language_config.runtime_version
        assert fetched_proc.language_config.handler == language_config.handler
        assert fetched_proc.language_config.packages == language_config.packages
        assert fetched_proc.language_config.imports == language_config.imports
        assert fetched_proc.execute_as == "OWNER"
        assert fetched_proc.comment is None
        assert fetched_proc.body is None
        assert fetched_proc.schema_name == procedures.schema.name.upper()
        assert fetched_proc.database_name == procedures.database.name.upper()
    finally:
        procedures[f"{proc_name}()"].drop(if_exists=True)


def test_create_python_proc(procedures, anaconda_package_available):
    del anaconda_package_available
    python_proc_no_args = Procedure(
        name="python_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                                return 'str'
                        """,
    )
    try:
        proc = procedures.create(python_proc_no_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_NO_ARGS"
    finally:
        proc.drop()

    python_proc_with_args = Procedure(
        name="python_proc_with_args",
        arguments=[
            Argument(name="arr", datatype="ARRAY", default=[1, 2, 3, 4]),
            Argument(name="vari", datatype="VARIANT", default={"str": 1, "test": 5}),
        ],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session, arr, vari):
                                print(arr)
                                print(vari["str"])
                        """,
    )
    try:
        proc = procedures.create(python_proc_with_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_WITH_ARGS"
        assert proc_fetch_result.arguments == [
            Argument(name="ARR", datatype="ARRAY", default=[1, 2, 3, 4]),
            Argument(name="VARI", datatype="VARIANT", default={"str": 1, "test": 5}),
        ]
        assert proc_fetch_result.return_type == ReturnDataType(datatype="VARCHAR", nullable=True)
        assert proc_fetch_result.language_config == PythonFunction(
            runtime_version="3.13",
            handler="handler",
            packages=["snowflake-snowpark-python"],
            imports=[],
            external_access_integrations=[],
            secrets={},
        )
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    python_proc_table_func = Procedure(
        name="python_proc_table_func",
        arguments=[
            Argument(name="num1", datatype="NUMBER"),
            Argument(name="varchar1", datatype="VARCHAR"),
            Argument(name="arr1", datatype="ARRAY"),
            Argument(name="object1", datatype="OBJECT"),
        ],
        return_type=ReturnTable(
            column_list=[
                ColumnType(name="col1", datatype="GEOGRAPHY"),
                ColumnType(name="col2", datatype="FLOAT"),
                ColumnType(name="col4", datatype="NUMBER"),
            ]
        ),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session, num1, varchar1, arr1, object1):
                                print(arr1)
                                return None
                        """,
    )
    try:
        proc = procedures.create(python_proc_table_func, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_TABLE_FUNC"
        assert proc_fetch_result.arguments == [
            Argument(name="NUM1", datatype="NUMBER"),
            Argument(name="VARCHAR1", datatype="VARCHAR"),
            Argument(name="ARR1", datatype="ARRAY"),
            Argument(name="OBJECT1", datatype="OBJECT"),
        ]
        assert proc_fetch_result.return_type == ReturnTable(
            column_list=[
                ColumnType(name="COL1", datatype="GEOGRAPHY"),
                ColumnType(name="COL2", datatype="FLOAT"),
                ColumnType(name="COL4", datatype="NUMBER"),
            ]
        )
        assert proc_fetch_result.language_config == PythonFunction(
            runtime_version="3.13",
            handler="handler",
            packages=["snowflake-snowpark-python"],
            imports=[],
            external_access_integrations=[],
            secrets={},
        )
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    python_proc_table_func = Procedure(
        name="python_proc_table_func",
        arguments=[Argument(name="symbol", datatype="VARCHAR")],
        return_type=ReturnTable(
            column_list=[ColumnType(name="symbol", datatype="VARCHAR"), ColumnType(name="price", datatype="NUMBER")]
        ),
        language_config=PythonFunction(runtime_version="3.13", handler="run", packages=["snowflake-snowpark-python"]),
        body="""from snowflake.snowpark import Session
from snowflake.snowpark.functions import col, sum

class StockSaleSumProc:
    def __init__(self, symbol):
        self._symbol = symbol

    def compute_total(self, session):
        df = session.table("stocks_table").filter(col("symbol") == self._symbol).select(col("symbol"),
                                                                        (col("quantity") * col("price")).alias("total")
                                                                                        )
        result_df = df.group_by(col("symbol")).agg(sum(col("total")).alias("total"))

        return result_df

def run(session: Session, symbol: str):
    stock_sale_sum = StockSaleSumProc(symbol)
    return stock_sale_sum.compute_total(session)
    """,
    )
    try:
        proc = procedures.create(python_proc_table_func, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_TABLE_FUNC"
        assert proc_fetch_result.arguments == [Argument(name="SYMBOL", datatype="VARCHAR")]
        assert proc_fetch_result.return_type == ReturnTable(
            column_list=[ColumnType(name="SYMBOL", datatype="VARCHAR"), ColumnType(name="PRICE", datatype="NUMBER")]
        )
        assert proc_fetch_result.language_config == PythonFunction(
            runtime_version="3.13",
            handler="run",
            packages=["snowflake-snowpark-python"],
            imports=[],
            external_access_integrations=[],
            secrets={},
        )
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()


def test_create_scala_proc(procedures, maven_snowpark_jar_available):
    del maven_snowpark_jar_available
    scala_proc_no_args = Procedure(
        name="scala_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=ScalaFunction(
            runtime_version="2.12", handler="TestScalaSP.asyncBasic", packages=["com.snowflake:snowpark:latest"]
        ),
        body="""object TestScalaSP {
                  def asyncBasic(session: com.snowflake.snowpark.Session): String = {
                    val df = session.sql("select system$wait(10)")
                    val asyncJob = df.async.collect()
                    while(!asyncJob.isDone()) {
                      Thread.sleep(1000)
                    }
                    "Done"
                  }
                }
            """,
    )
    try:
        proc = procedures.create(scala_proc_no_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "SCALA_PROC_NO_ARGS"
    finally:
        proc.drop()

    scala_proc_with_args = Procedure(
        name="scala_proc_with_args",
        arguments=[Argument(name="fileName", datatype="VARCHAR")],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=ScalaFunction(
            runtime_version="2.12", handler="FileReader.execute", packages=["com.snowflake:snowpark:latest"]
        ),
        body="""import java.io.InputStream
                import java.nio.charset.StandardCharsets
                import com.snowflake.snowpark_java.types.SnowflakeFile
                import com.snowflake.snowpark_java.Session

                object FileReader {
                  def execute(session: Session, fileName: String): String = {
                    var input: InputStream = SnowflakeFile.newInstance(fileName).getInputStream()
                    return new String(input.readAllBytes(), StandardCharsets.UTF_8)
                  }
                }
            """,
    )
    try:
        proc = procedures.create(scala_proc_with_args, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        scalafun = proc_fetch_result.language_config
        assert proc_fetch_result.name == "SCALA_PROC_WITH_ARGS"
        assert proc_fetch_result.arguments == [Argument(name="FILENAME", datatype="VARCHAR")]
        assert proc_fetch_result.return_type == ReturnDataType(datatype="VARCHAR", nullable=True)
        assert scalafun.runtime_version == "2.12"
        assert scalafun.handler == "FileReader.execute"
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()

    scala_proc_table_func = Procedure(
        name="scala_proc_table_func",
        arguments=[Argument(name="tablename", datatype="VARCHAR"), Argument(name="role", datatype="VARCHAR")],
        comment="scala proc",
        is_secure=True,
        return_type=ReturnTable(
            column_list=[
                ColumnType(name="id", datatype="NUMBER"),
                ColumnType(name="name", datatype="VARCHAR"),
                ColumnType(name="role", datatype="VARCHAR"),
            ]
        ),
        language_config=ScalaFunction(
            runtime_version="2.12", handler="Filter.filterByRole", packages=["com.snowflake:snowpark:latest"]
        ),
        body="""import com.snowflake.snowpark.functions._
                import com.snowflake.snowpark._

                object Filter {
                   def filterByRole(session: Session, tableName: String, role: String): DataFrame = {
                     val table = session.table(tableName)
                     val filteredRows = table.filter(col("role") === role)
                     return filteredRows
                   }
                }
            """,
    )
    try:
        proc = procedures.create(scala_proc_table_func, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        scalafun = proc_fetch_result.language_config
        assert proc_fetch_result.name == "SCALA_PROC_TABLE_FUNC"
        assert proc_fetch_result.arguments == [
            Argument(name="TABLENAME", datatype="VARCHAR"),
            Argument(name="ROLE", datatype="VARCHAR"),
        ]
        assert proc_fetch_result.return_type == ReturnTable(
            column_list=[
                ColumnType(name="ID", datatype="NUMBER"),
                ColumnType(name="NAME", datatype="VARCHAR"),
                ColumnType(name="ROLE", datatype="VARCHAR"),
            ]
        )
        assert scalafun.runtime_version == "2.12"
        assert scalafun.handler == "Filter.filterByRole"
        assert proc_fetch_result.comment == "scala proc"
        assert proc_fetch_result.is_secure is True
        assert proc_fetch_result.is_builtin is True
    finally:
        proc.drop()


@pytest.mark.usefixtures("anaconda_package_available")
def test_create_procedure_or_replace(procedures):
    python_proc_no_args1 = Procedure(
        name="python_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                    return 'str'
            """,
    )
    python_proc_no_args2 = Procedure(
        name="python_proc_no_args",
        arguments=[],
        return_type=ReturnDataType(datatype="NUMBER"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                            return 'str'
                    """,
    )
    try:
        proc = procedures.create(python_proc_no_args1, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_NO_ARGS"
        assert proc_fetch_result.return_type.datatype == "VARCHAR"

        proc = procedures.create(python_proc_no_args2, mode=CreateMode.or_replace)
        proc_fetch_result = proc.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_NO_ARGS"
        assert proc_fetch_result.return_type.datatype == "NUMBER"
    finally:
        proc.drop()


@pytest.mark.usefixtures("anaconda_package_available")
def test_create_procedure_error_if_exists(procedures):
    python_proc_no_args1 = Procedure(
        name="python_proc_no_args1",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                    return 'str'
            """,
    )

    python_proc_no_args2 = Procedure(
        name="python_proc_no_args1",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                        return 'str'
                """,
    )
    try:
        proc1 = procedures.create(python_proc_no_args1)
        with pytest.raises(ConflictError):
            procedures.create(python_proc_no_args2, mode=CreateMode.error_if_exists)
    finally:
        proc1.drop()


@pytest.mark.usefixtures("anaconda_package_available")
def test_create_procedure_if_not_exists(procedures):
    python_proc_no_args1 = Procedure(
        name="python_proc_no_args2",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                        return 'str'
                """,
    )

    python_proc_no_args2 = Procedure(
        name="python_proc_no_args2",
        arguments=[],
        return_type=ReturnDataType(datatype="NUMBER"),
        language_config=PythonFunction(
            runtime_version="3.13", handler="handler", packages=["snowflake-snowpark-python"]
        ),
        body="""def handler(session):
                            return 'str'
                    """,
    )
    try:
        proc1 = procedures.create(python_proc_no_args1, mode=CreateMode.or_replace)
        proc_fetch_result = proc1.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_NO_ARGS2"
        assert proc_fetch_result.return_type.datatype == "VARCHAR"

        proc2 = procedures.create(python_proc_no_args2, mode=CreateMode.if_not_exists)
        proc_fetch_result = proc2.fetch()
        assert proc_fetch_result.name == "PYTHON_PROC_NO_ARGS2"
        assert proc_fetch_result.return_type.datatype == "VARCHAR"
    finally:
        proc1.drop()


def test_create_procedure_invalid_language(procedures):
    random_proc = Procedure(
        name="random_proc",
        arguments=[],
        return_type=ReturnDataType(datatype="VARCHAR"),
        language_config=FunctionLanguage(),
        body="""return""",
    )
    with pytest.raises(APIError):
        procedures.create(random_proc, mode=CreateMode.error_if_exists)



from tests.integ.fixtures.temp_objects_for_grant import (
    database_roles_in_test_database,  # noqa: F401
    schemas_in_test_database,  # noqa: F401
    tables_in_test_database,  # noqa: F401
    test_database_for_grant_name,  # noqa: F401
    test_database_role_name,  # noqa: F401
    test_database_role_name_2,  # noqa: F401
    test_role_name,  # noqa: F401
    test_role_name_2,  # noqa: F401
    test_schema_for_grant_name,  # noqa: F401
    test_table_for_grant_name,  # noqa: F401
)



"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any

from pydantic import BaseModel, ConfigDict, StrictStr


class OpenAIReasoning(BaseModel):
    """A model object representing the OpenAIReasoning resource.

    Constructs an object of type OpenAIReasoning with the provided properties.

    Parameters
    __________
    summary : str
        Reasoning summary tokens generated during the model response.
    encrypted_content : str
        The encrypted_content field is an opaque field and should not be interpreted or parsed.
        It should only be used for verifying that the reasoning blocks were generated by that specific model.
    """

    summary: StrictStr

    encrypted_content: StrictStr

    __properties = ["summary", "encrypted_content"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> OpenAIReasoning:
        """Create an instance of OpenAIReasoning from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> OpenAIReasoning:
        """Create an instance of OpenAIReasoning from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return OpenAIReasoning.model_validate(obj)

        _obj = OpenAIReasoning.model_validate(
            {
                "summary": obj.get("summary"),
                "encrypted_content": obj.get("encrypted_content"),
            }
        )

        return _obj


class OpenAIReasoningModel:
    def __init__(
        self,
        summary: str,
        encrypted_content: str,
        # optional properties
    ):
        """A model object representing the OpenAIReasoning resource.

                Constructs an object of type OpenAIReasoning with the provided properties.

                Parameters
                __________
                summary : str
                    Reasoning summary tokens generated during the model response.

                encrypted_content : str
                    The encrypted_content field is an opaque field and should not be interpreted or parsed.
        It should only be used for verifying that the reasoning blocks were generated by that specific model.
        """
        self.summary = summary
        self.encrypted_content = encrypted_content

    __properties = ["summary", "encrypted_content"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return OpenAIReasoning(
            summary=self.summary,
            encrypted_content=self.encrypted_content,
        )

    @classmethod
    def _from_model(cls, model) -> OpenAIReasoningModel:
        return OpenAIReasoningModel(
            summary=model.summary,
            encrypted_content=model.encrypted_content,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a OpenAIReasoning.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the OpenAIReasoning object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> OpenAIReasoningModel:
        """Create an instance of OpenAIReasoning from a dict.

        This method constructs a OpenAIReasoning object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        OpenAIReasoning
            A OpenAIReasoning object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(OpenAIReasoning.from_dict(obj))


OpenAIReasoning._model_class = OpenAIReasoningModel



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.user import Securable


pytestmark = pytest.mark.min_sf_ver("8.39.0")


@pytest.mark.use_accountadmin
def test_grant_user(users, test_user_name, test_role_name):
    users[test_user_name].grant_role(role_type="ROLE", role=Securable(name=test_role_name))

    assert len(list(users[test_user_name].iter_grants_to())) == 1

    # running the same grant again should not be an issue
    users[test_user_name].grant_role(role_type="ROLE", role=Securable(name=test_role_name))

    assert len(list(users[test_user_name].iter_grants_to())) == 1

    # granting a RANDOM ROLE which does not exist
    with pytest.raises(NotFoundError):
        users[test_user_name].grant_role(role_type="ROLE", role=Securable(name="random_role"))

    # granting a ROLE to a user which does not exist
    with pytest.raises(NotFoundError):
        users["RANDOM"].grant_role(role_type="ROLE", role=Securable(name=test_role_name))



from __future__ import annotations


class Grantee:
    """
    Class to represent type of resource that is the privilege grantee.

    Examples
    ________
    Setting test-user to user and test-role to role:

    >>> Grantee("test-user", "user")
    >>> Grantee("test-role", "role")
    """

    def __init__(self, name: str, grantee_type: str):
        self._name = name
        self._grantee_type = grantee_type

    @property
    def name(self) -> str:
        """String that specifies the name of the privilege grantee."""
        return self._name

    @property
    def grantee_type(self) -> str:
        """String that specifies the type of resource that is the privilege grantee."""
        return self._grantee_type


class Grantees:
    """
    Util Class with static methods to create various :class:`Grantee` class instances.

    Examples
    ________
    Setting test-user to user and test-role to role:

    >>> Grantees.role("test-role")
    >>> Grantees.user("test-user")
    """

    @staticmethod
    def user(name: str) -> Grantee:
        return Grantee(name, "user")

    @staticmethod
    def role(name: str) -> Grantee:
        return Grantee(name, "role")

    @staticmethod
    def database_role(name: str) -> Grantee:
        return Grantee(name, "database-role")

    @staticmethod
    def share(name: str) -> Grantee:
        return Grantee(name, "share")

    @staticmethod
    def application_role(name: str) -> Grantee:
        return Grantee(name, "application-role")

    @staticmethod
    def application(name: str) -> Grantee:
        return Grantee(name, "application")



# coding: utf-8

# flake8: noqa
"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

# import models into model package
from snowflake.core.cortex.analyst_service._generated.models.agentic_optimization_status import (
    AgenticOptimizationStatus,
)
from snowflake.core.cortex.analyst_service._generated.models.agentic_optimization_summary import (
    AgenticOptimizationSummary,
)
from snowflake.core.cortex.analyst_service._generated.models.annotated_sql_query import AnnotatedSqlQuery
from snowflake.core.cortex.analyst_service._generated.models.confidence import Confidence
from snowflake.core.cortex.analyst_service._generated.models.create_agentic_optimization_request import (
    CreateAgenticOptimizationRequest,
)
from snowflake.core.cortex.analyst_service._generated.models.create_agentic_optimization_response import (
    CreateAgenticOptimizationResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.create_analyst_eval_optimization_request import (
    CreateAnalystEvalOptimizationRequest,
)
from snowflake.core.cortex.analyst_service._generated.models.create_analyst_eval_optimization_response import (
    CreateAnalystEvalOptimizationResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.error_response import ErrorResponse
from snowflake.core.cortex.analyst_service._generated.models.fast_generation_request import FastGenerationRequest
from snowflake.core.cortex.analyst_service._generated.models.fast_generation_response import FastGenerationResponse
from snowflake.core.cortex.analyst_service._generated.models.generate_filters_and_metrics_suggestions_request import (
    GenerateFiltersAndMetricsSuggestionsRequest,
)
from snowflake.core.cortex.analyst_service._generated.models.generate_filters_and_metrics_suggestions_response import (
    GenerateFiltersAndMetricsSuggestionsResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.generate_verified_query_suggestions_request import (
    GenerateVerifiedQuerySuggestionsRequest,
)
from snowflake.core.cortex.analyst_service._generated.models.generate_verified_query_suggestions_response import (
    GenerateVerifiedQuerySuggestionsResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.get_agentic_optimization_response import (
    GetAgenticOptimizationResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.list_agentic_optimizations_request import (
    ListAgenticOptimizationsRequest,
)
from snowflake.core.cortex.analyst_service._generated.models.list_agentic_optimizations_response import (
    ListAgenticOptimizationsResponse,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content import MessageContent
from snowflake.core.cortex.analyst_service._generated.models.message_content_chart_object import (
    MessageContentChartObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_delta import MessageContentDelta
from snowflake.core.cortex.analyst_service._generated.models.message_content_delta_sql_object import (
    MessageContentDeltaSqlObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_delta_suggestions_object import (
    MessageContentDeltaSuggestionsObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_delta_text_object import (
    MessageContentDeltaTextObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_delta_think_object import (
    MessageContentDeltaThinkObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_results_object import (
    MessageContentResultsObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_sql_object import MessageContentSqlObject
from snowflake.core.cortex.analyst_service._generated.models.message_content_suggestions_object import (
    MessageContentSuggestionsObject,
)
from snowflake.core.cortex.analyst_service._generated.models.message_content_text_object import MessageContentTextObject
from snowflake.core.cortex.analyst_service._generated.models.message_object import MessageObject
from snowflake.core.cortex.analyst_service._generated.models.pre_selection_request import PreSelectionRequest
from snowflake.core.cortex.analyst_service._generated.models.pre_selection_response import PreSelectionResponse
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_change import SemanticModelChange
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_change_operation import (
    SemanticModelChangeOperation,
)
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_object import SemanticModelObject
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_selection import SemanticModelSelection
from snowflake.core.cortex.analyst_service._generated.models.semantic_model_update_suggestion import (
    SemanticModelUpdateSuggestion,
)
from snowflake.core.cortex.analyst_service._generated.models.send_feedback_request import SendFeedbackRequest
from snowflake.core.cortex.analyst_service._generated.models.send_message_request import SendMessageRequest
from snowflake.core.cortex.analyst_service._generated.models.send_message_response import SendMessageResponse
from snowflake.core.cortex.analyst_service._generated.models.sql_inputs import SqlInputs
from snowflake.core.cortex.analyst_service._generated.models.staging_data import StagingData
from snowflake.core.cortex.analyst_service._generated.models.status_update import StatusUpdate
from snowflake.core.cortex.analyst_service._generated.models.streaming_error import StreamingError
from snowflake.core.cortex.analyst_service._generated.models.success_response import SuccessResponse
from snowflake.core.cortex.analyst_service._generated.models.suggestion_delta import SuggestionDelta
from snowflake.core.cortex.analyst_service._generated.models.table_columns import TableColumns
from snowflake.core.cortex.analyst_service._generated.models.tableau_auth import TableauAuth
from snowflake.core.cortex.analyst_service._generated.models.tableau_input import TableauInput
from snowflake.core.cortex.analyst_service._generated.models.tableau_integration import TableauIntegration
from snowflake.core.cortex.analyst_service._generated.models.tableau_server_files import TableauServerFiles
from snowflake.core.cortex.analyst_service._generated.models.tableau_stage_files import TableauStageFiles
from snowflake.core.cortex.analyst_service._generated.models.tds_input import TdsInput
from snowflake.core.cortex.analyst_service._generated.models.token_info import TokenInfo
from snowflake.core.cortex.analyst_service._generated.models.verified_query import VerifiedQuery
from snowflake.core.cortex.analyst_service._generated.models.verified_query_suggestion import VerifiedQuerySuggestion
from snowflake.core.cortex.analyst_service._generated.models.warning import Warning
from snowflake.core.cortex.analyst_service._generated.models.warnings import Warnings

__all__ = [
    "AgenticOptimizationStatus",
    "AgenticOptimizationSummary",
    "AnnotatedSqlQuery",
    "Confidence",
    "CreateAgenticOptimizationRequest",
    "CreateAgenticOptimizationResponse",
    "CreateAnalystEvalOptimizationRequest",
    "CreateAnalystEvalOptimizationResponse",
    "ErrorResponse",
    "FastGenerationRequest",
    "FastGenerationResponse",
    "GenerateFiltersAndMetricsSuggestionsRequest",
    "GenerateFiltersAndMetricsSuggestionsResponse",
    "GenerateVerifiedQuerySuggestionsRequest",
    "GenerateVerifiedQuerySuggestionsResponse",
    "GetAgenticOptimizationResponse",
    "ListAgenticOptimizationsRequest",
    "ListAgenticOptimizationsResponse",
    "MessageContent",
    "MessageContentChartObject",
    "MessageContentDelta",
    "MessageContentDeltaSqlObject",
    "MessageContentDeltaSuggestionsObject",
    "MessageContentDeltaTextObject",
    "MessageContentDeltaThinkObject",
    "MessageContentResultsObject",
    "MessageContentSqlObject",
    "MessageContentSuggestionsObject",
    "MessageContentTextObject",
    "MessageObject",
    "PreSelectionRequest",
    "PreSelectionResponse",
    "SemanticModelChange",
    "SemanticModelChangeOperation",
    "SemanticModelObject",
    "SemanticModelSelection",
    "SemanticModelUpdateSuggestion",
    "SendFeedbackRequest",
    "SendMessageRequest",
    "SendMessageResponse",
    "SqlInputs",
    "StagingData",
    "StatusUpdate",
    "StreamingError",
    "SuccessResponse",
    "SuggestionDelta",
    "TableColumns",
    "TableauAuth",
    "TableauInput",
    "TableauIntegration",
    "TableauServerFiles",
    "TableauStageFiles",
    "TdsInput",
    "TokenInfo",
    "VerifiedQuery",
    "VerifiedQuerySuggestion",
    "Warning",
    "Warnings",
]



from typing import TYPE_CHECKING

from ._generated.api.password_policy_api_base import PasswordPolicyCollectionBase, PasswordPolicyResourceBase


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class PasswordPolicyCollection(PasswordPolicyCollectionBase):
    """Represents the collection operations on the Snowflake password policy resource.

    With this collection, you can create, iterate through, and search for password policies that you have access to
    in the current context.
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, PasswordPolicyResource)


class PasswordPolicyResource(PasswordPolicyResourceBase):
    """Represents a reference to a Snowflake password policy.

    With this password policy reference, you can fetch information about a password policy, as well as
    perform certain actions on it.
    """

    _plural_name = "password_policies"



from collections.abc import Iterator
from concurrent.futures import Future
from functools import cached_property
from typing import TYPE_CHECKING, Literal, Optional, Union, overload

from pydantic import StrictStr

from snowflake.core import PollingOperation
from snowflake.core._common import AccountObjectCollectionParent, Clone, CreateMode, ObjectReferenceMixin, PointOfTime
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._internal.utils import deprecated
from snowflake.core._operation import PollingOperations
from snowflake.core._utils import (
    tag_assignment_to_tag_tuple,
    tag_resource_to_tag_reference,
    tag_tuple_to_tag_assignment,
)
from snowflake.core.database._generated import SuccessResponse
from snowflake.core.database._generated.api import DatabaseApi
from snowflake.core.database._generated.models.account_identifiers import AccountIdentifiers
from snowflake.core.database._generated.models.database import DatabaseModel as Database
from snowflake.core.database._generated.models.database_clone import DatabaseClone
from snowflake.core.database._generated.models.database_from_share import DatabaseFromShare
from snowflake.core.database._generated.models.point_of_time import PointOfTime as DatabasePointOfTime
from snowflake.core.database._generated.models.tag_assignment import TagAssignment
from snowflake.core.database._generated.models.tag_reference import TagReference
from snowflake.core.database_role import DatabaseRoleCollection
from snowflake.core.schema import SchemaCollection
from snowflake.core.tag import TagResource, TagValue


if TYPE_CHECKING:
    from snowflake.core import Root


class DatabaseCollection(AccountObjectCollectionParent["DatabaseResource"]):
    """Represents the collection operations on the Snowflake Database resource.

    With this collection, you can create, iterate through, and search for Databases that you have access to in the
    current context.

    Examples
    ________
    Creating a database instance:

    >>> databases = root.databases
    >>> new_database = Database(
    ...     name="my_new_database", comment="this is my new database to prototype a new feature in"
    ... )
    >>> databases.create(new_database)
    """

    def __init__(self, root: "Root") -> None:
        super().__init__(root, ref_class=DatabaseResource)
        self._api = DatabaseApi(
            root=root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(
        self,
        database: Database,
        *,
        clone: Optional[Union[str, Clone]] = None,
        from_share: Optional[str] = None,
        mode: CreateMode = CreateMode.error_if_exists,
    ) -> "DatabaseResource":
        """Create a database in Snowflake.

        Parameters
        __________
        database: Database
            The ``Database`` object, together with the ``Database``'s properties: name; kind, comment, retention_time,
            budget, data_retention_time_in_days, default_ddl_collation, log_level, max_data_extension_time_in_days,
            suspend_task_after_num_failures, trace_level, user_task_managed_initial_warehouse_size, user_task_timeout_ms
            serverless_task_min_statement_size and serverless_task_max_statement_size are optional.
        clone: str, or Clone, optional
            Whether to clone an existing database. An instance of :class:`Clone`, or ``None``
            if no cloning is necessary.
        from_share: str, optional
            ID of the share from which to create the database, in the form ``"<provider_account>.<share_name>``".
        mode: CreateMode, optional
            One of the following enum values.

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the database already exists in Snowflake.  Equivalent to SQL ``create database <name> ...``.

            ``CreateMode.or_replace``: Replace if the database already exists in Snowflake. Equivalent to SQL
            ``create or replace database <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the database already exists in Snowflake.
            Equivalent to SQL ``create database <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.

            Example
            _______
            Creating a new database if it does not already exist:

            >>> new_db_ref = root.databases.create(Database(name="my_new_database"), mode=CreateMode.if_not_exists)
            >>> print(new_db_ref.fetch())
        """
        self._create(database=database, clone=clone, from_share=from_share, mode=mode, async_req=False)
        return self[database.name]

    @api_telemetry
    def create_async(
        self,
        database: Database,
        *,
        clone: Optional[Union[str, Clone]] = None,
        from_share: Optional[str] = None,
        mode: CreateMode = CreateMode.error_if_exists,
    ) -> PollingOperation["DatabaseResource"]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._create(database=database, clone=clone, from_share=from_share, mode=mode, async_req=True)
        return PollingOperation(future, lambda _: self[database.name])

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
    ) -> Iterator[Database]:
        """
        Iterate through ``Database`` objects from Snowflake,  filtering on any optional 'like' pattern.

        Parameters
        _________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (``%`` and ``_``).
        starts_with: str, optional
            String used to filter the command output based on the string of characters that appear
            at the beginning of the object name. Uses case-sensitive pattern matching.
        limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.
        from_name: str, optional
            Fetch rows only following the first row whose object name matches
            the specified string. This is case-sensitive and does not have to be the full name.

        Examples
        ________
        Showing all databases that you have access to see:

        >>> databases = root.databases.iter()

        Showing information of the exact database you want to see:

        >>> databases = root.databases.iter(like="your-database-name")

        Showing databases starting with 'your-database-name-':

        >>> databases = root.databases.iter(like="your-database-name-%")

        Using a for loop to retrieve information from iterator:

        >>> for database in databases:
        >>>     print(database.name, database.query)
        """
        databases = self._api.list_databases(
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name=from_name,
            async_req=False,
        )

        return map(Database._from_model, iter(databases))

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
    ) -> PollingOperation[Iterator[Database]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.list_databases(
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name=from_name,
            async_req=True,
        )
        return PollingOperation(future, lambda rest_models: map(Database._from_model, iter(rest_models)))

    @overload
    def _create(
        self,
        database: Database,
        clone: Optional[Union[str, Clone]],
        from_share: Optional[str],
        mode: CreateMode,
        async_req: Literal[True],
    ) -> Future[SuccessResponse]: ...

    @overload
    def _create(
        self,
        database: Database,
        clone: Optional[Union[str, Clone]],
        from_share: Optional[str],
        mode: CreateMode,
        async_req: Literal[False],
    ) -> SuccessResponse: ...

    def _create(
        self,
        database: Database,
        clone: Optional[Union[str, Clone]],
        from_share: Optional[str],
        mode: CreateMode,
        async_req: bool,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        if clone is not None and from_share is not None:
            raise ValueError("Cannot use both `clone` and `from_share`")

        real_mode = CreateMode[mode].value
        if clone is not None:
            pot: Optional[DatabasePointOfTime] = None
            if isinstance(clone, Clone) and isinstance(clone.point_of_time, PointOfTime):
                pot = DatabasePointOfTime.from_dict(clone.point_of_time.to_dict())
            real_clone = Clone(source=clone) if isinstance(clone, str) else clone
            clone_req = DatabaseClone(point_of_time=pot, **database._to_model().to_dict())
            return self._api.clone_database(
                name=real_clone.source, database_clone=clone_req, create_mode=StrictStr(real_mode), async_req=async_req
            )

        if from_share is not None:
            share_req = DatabaseFromShare(**database._to_model().to_dict())
            return self._api.create_database_from_share(
                database_from_share=share_req, share=from_share, create_mode=StrictStr(real_mode), async_req=async_req
            )

        return self._api.create_database(
            database=database._to_model(), create_mode=StrictStr(real_mode), async_req=async_req
        )


class DatabaseResource(ObjectReferenceMixin[DatabaseCollection]):
    """Represents a reference to a Snowflake database.

    With this database reference, you can fetch information about a database, as well as perform
    certain actions on it.
    """

    def __init__(self, name: str, collection: DatabaseCollection) -> None:
        self.name = name
        self.collection: DatabaseCollection = collection

    @property
    def _api(self) -> DatabaseApi:
        return self.collection._api

    @api_telemetry
    def fetch(self) -> Database:
        """Fetch the details of a database.

        Examples
        ________
        Fetching a reference to a database to print whether it's the currently used database:

        >>> my_database = root.databases["my_db"].fetch()
        >>> print(my_database.is_current)
        """
        return Database._from_model(self.collection._api.fetch_database(self.name, async_req=False))

    @api_telemetry
    def fetch_async(self) -> PollingOperation[Database]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_database(self.name, async_req=True)
        return PollingOperation(future, lambda rest_model: Database._from_model(rest_model))

    @api_telemetry
    @deprecated("drop")
    def delete(self, if_exists: Optional[bool] = None) -> None:
        """Delete this database.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this database before dropping it.
            The default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Deleting a database using its reference:

        >>> root.databases["to_be_deleted"].delete()
        """
        self.drop(if_exists=if_exists)

    @api_telemetry
    def drop(self, if_exists: Optional[bool] = None) -> None:
        """Drop this database.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this database before dropping it.
            The default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Using a database reference to drop a database, erroring if it does not exist:

        >>> root.databases["to_be_dropped"].drop()

        Using a database reference to drop a database, if it exists:

        >>> root.databases["to_be_dropped"].drop(if_exists=True)
        """
        self.collection._api.delete_database(name=self.name, if_exists=if_exists, async_req=False)

    @api_telemetry
    def drop_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.delete_database(name=self.name, if_exists=if_exists, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def undrop(self) -> None:
        """Undrop this database.

        Examples
        ________
        Undropping a database using its reference:

        >>> root.databases["to_be_undropped"].undrop()
        """
        self.collection._api.undrop_database(name=self.name, async_req=False)

    @api_telemetry
    def undrop_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`undrop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.undrop_database(name=self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    @deprecated("create_or_alter")
    def create_or_update(self, database: Database) -> None:
        """Create or update an existing database.

        Examples
        ________
        Create a database from a reference:

        >>> root.databases["my_new_db"].create_or_update(Database("my_new_db"))

        """
        self.create_or_alter(database=database)

    @api_telemetry
    def create_or_alter(self, database: Database) -> None:
        """Create or alter a database in Snowflake."""
        self._api.create_or_alter_database(database.name, database._to_model(), async_req=False)

    @api_telemetry
    def create_or_alter_async(self, database: Database) -> PollingOperation[None]:
        """An asynchronous version of :func:`create_or_alter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.create_or_alter_database(database.name, database._to_model(), async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def enable_replication(self, accounts: list[str], ignore_edition_check: bool = False) -> None:
        """Promotes a local database to serve as a primary database for replication.

        A primary database can be replicated in one or more accounts, allowing users
        in those accounts to query objects in each secondary (i.e. replica) database.

        Alternatively, modify an existing primary database to add to or remove from
        the list of accounts that can store a replica of the database.

        Provide a list of accounts in your organization that can store a replica of
        this database.

        Parameters
        __________
        accounts: list of str
            Array of unique account identifiers for which to enable replication.
        ignore_edition_check: bool, optional
            Whether to allow replicating data to accounts on lower editions.

            Default is ``True``.

        Examples
        ________
        Enabling replication of "my_db" database on 2 other accounts using its reference:

        >>> root.databases["my_db"].enable_replication(
        ...     accounts=["accountName1", "accountName2"], ignore_edition_check=True
        ... )
        """
        if len(accounts) == 0:
            raise ValueError("Account list given to replication cannot be empty.")
        self.collection._api.enable_database_replication(
            name=self.name,
            account_identifiers=AccountIdentifiers(accounts=accounts),
            ignore_edition_check=ignore_edition_check,
            async_req=False,
        )

    @api_telemetry
    def enable_replication_async(
        self, accounts: list[str], ignore_edition_check: bool = False
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`enable_replication`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        if len(accounts) == 0:
            raise ValueError("Account list given to replication cannot be empty.")
        future = self.collection._api.enable_database_replication(
            name=self.name,
            account_identifiers=AccountIdentifiers(accounts=accounts),
            ignore_edition_check=ignore_edition_check,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def disable_replication(self, accounts: Optional[list[str]] = None) -> None:
        """Disables replication for this primary database.

        Which means that no replica of this database (i.e. secondary database) in
        another account can be refreshed. Any secondary databases remain linked to
        the primary database, but requests to refresh a secondary database are denied.

        Note that disabling replication for a primary database does not prevent it
        from being replicated to the same account; therefore, the database continues
        to be listed in the SHOW REPLICATION DATABASES output.

        Optionally provide a comma-separated list of accounts in your organization
        to disable replication for this database only in the specified accounts.

        Parameters
        __________
        accounts: list of str, optional
            Array of unique account identifiers for which to disable replication.

        Examples
        ________
        Disabling all replication of "my_db" database using its reference:

        >>> root.databases["my_db"].disable_replication()
        """
        if accounts is None:
            accounts = []
        self.collection._api.disable_database_replication(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=False
        )

    @api_telemetry
    def disable_replication_async(self, accounts: Optional[list[str]] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`disable_replication`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        if accounts is None:
            accounts = []
        future = self.collection._api.disable_database_replication(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def refresh_replication(self) -> None:
        """Refresh a secondary database from its primary database.

        A snapshot includes changes to the objects and data.

        Examples
        ________
        Refreshing a database replication using its reference:

        >>> root.databases["db_replication"].refresh_replication()
        """
        self.collection._api.refresh_database_replication(name=self.name, async_req=False)

    @api_telemetry
    def refresh_replication_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`refresh_replication`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.refresh_database_replication(name=self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def enable_failover(self, accounts: list[str]) -> None:
        """Enable a list of replicas of this database that can be promoted to primary.

        Parameters
        __________
        accounts: list of str
            Array of unique account identifiers for which to enable failover.

        Examples
        ________
        Enabling failover to an account using a database reference:

        >>> root.databases["my_db"].enable_failover(accounts=["my_failover_acc"])
        """
        if len(accounts) == 0:
            raise ValueError("Account list given to replication cannot be empty.")
        self.collection._api.enable_database_failover(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=False
        )

    @api_telemetry
    def enable_failover_async(self, accounts: list[str]) -> PollingOperation[None]:
        """An asynchronous version of :func:`enable_failover`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        if len(accounts) == 0:
            raise ValueError("Account list given to replication cannot be empty.")
        future = self.collection._api.enable_database_failover(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def disable_failover(self, accounts: Optional[list[str]] = None) -> None:
        """Disables failover for this primary databases.

        Which means that no replica of this database (i.e. secondary database) can be
        promoted to serve as the primary database.

        Optionally provide a comma-separated list of accounts in your organization to
        disable failover for this database only in the specified accounts.

        Parameters
        __________
        accounts: list of str, optional
            Array of unique account identifiers for which to enable failover.

        Examples
        ________
        Disabling failover to some account using a database reference:

        >>> root.databases["my_db"].enable_failover(accounts=["old_failover_acc"])
        """
        if accounts is None:
            accounts = []
        self.collection._api.disable_database_failover(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=False
        )

    @api_telemetry
    def disable_failover_async(self, accounts: Optional[list[str]] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`disable_failover`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        if accounts is None:
            accounts = []
        future = self.collection._api.disable_database_failover(
            name=self.name, account_identifiers=AccountIdentifiers(accounts=accounts), async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def promote_to_primary_failover(self) -> None:
        """Promotes the specified secondary (replica) database to serve as the primary.

        When promoted, the database becomes writeable. At the same time, the previous
        primary database becomes a read-only secondary database.

        Examples
        ________
        Promoting a ``my_db`` failover database to be the primary using its reference:

        >>> root.databases["my_db"].promote_to_primary_failover()
        """
        self.collection._api.primary_database_failover(name=self.name, async_req=False)

    @api_telemetry
    def promote_to_primary_failover_async(self) -> PollingOperation[None]:
        """An asynchronous version of :func:`promote_to_primary_failover`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.primary_database_failover(name=self.name, async_req=True)
        return PollingOperations.empty(future)

    @api_telemetry
    def set_tags(self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None) -> None:
        """Set tags on a database.

        Parameters
        __________
        tags: dict[TagResource, TagValue]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.set_tags(
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
        )

    @api_telemetry
    def set_tags_async(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`set_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.set_tags(
            name=self.name,
            tag_assignment=[
                tag_tuple_to_tag_assignment(TagAssignment, tag_resource, tag_value)
                for [tag_resource, tag_value] in tags.items()
            ],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def unset_tags(self, tag_resources: set[TagResource], if_exists: Optional[bool] = None) -> None:
        """Unset tags from a database.

        Parameters
        __________
        tag_resources: set[TagResource]
             (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`:
             The endpoint does not throw an error if the resource does not exist. It returns a 200 success response,
             but does not take any action on the resource. - `false`: The endpoint throws an error if the resource
             doesn't exist.
        """
        self.collection._api.unset_tags(
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
        )

    @api_telemetry
    def unset_tags_async(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = None
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`unset_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.unset_tags(
            name=self.name,
            tag_reference=[tag_resource_to_tag_reference(TagReference, tag_resource) for tag_resource in tag_resources],
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def get_tags(self, with_lineage: Optional[bool] = None) -> dict[TagResource, TagValue]:
        """Get the tag assignments for a database.

        Returns all tags assigned to a database. This operation requires an active warehouse.

        Parameters
        __________
        with_lineage: bool, optional
            Parameter that specifies whether tag assignments inherited by the object from its ancestors in securable
            object hierarchy should be returned as well: - `true`: All tags assigned to this object should be returned,
            inheritance included. - `false`: Only tags explicitly assigned to this object should be returned.
        """
        tag_assignments = self.collection._api.get_tags(
            name=self.name,
            with_lineage=with_lineage,
        )
        return dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)

    @api_telemetry
    def get_tags_async(self, with_lineage: Optional[bool] = None) -> PollingOperation[dict[TagResource, TagValue]]:
        """An asynchronous version of :func:`get_tags`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.get_tags(name=self.name, with_lineage=with_lineage, async_req=True)
        return PollingOperation(
            future, lambda tag_assignments: dict(tag_assignment_to_tag_tuple(ta, self.root) for ta in tag_assignments)
        )

    @cached_property
    def schemas(self) -> SchemaCollection:
        """The SchemaCollection of all schemas contained in this database.

        Examples
        ________
        Getting all schemas in ``my_db``:

        >>> root.databases["my_db"].schemas
        """
        return SchemaCollection(self, self.root)

    @cached_property
    def database_roles(self) -> DatabaseRoleCollection:
        """The DatabaseRoleCollection of all database roles contained in this database.

        Examples
        ________
        Getting all database roles in ``my_db``:

        >>> root.databases["my_db"].database_roles
        """
        return DatabaseRoleCollection(self, self.root)



# coding: utf-8

# flake8: noqa
"""
Cortex Search REST API.

OpenAPI 3.0 specification for the Cortex Search REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

# import models into model package
from snowflake.core.cortex.search_service._generated.models.column_query import ColumnQuery
from snowflake.core.cortex.search_service._generated.models.cortex_search_service import CortexSearchService
from snowflake.core.cortex.search_service._generated.models.create_cortex_search_service_request import (
    CreateCortexSearchServiceRequest,
)
from snowflake.core.cortex.search_service._generated.models.diversity import Diversity
from snowflake.core.cortex.search_service._generated.models.downstream_lag import DownstreamLag
from snowflake.core.cortex.search_service._generated.models.error_response import ErrorResponse
from snowflake.core.cortex.search_service._generated.models.functions import Functions
from snowflake.core.cortex.search_service._generated.models.numeric_boost import NumericBoost
from snowflake.core.cortex.search_service._generated.models.query_request import QueryRequest
from snowflake.core.cortex.search_service._generated.models.query_response import QueryResponse
from snowflake.core.cortex.search_service._generated.models.query_similarity_boost import QuerySimilarityBoost
from snowflake.core.cortex.search_service._generated.models.ranking_weights import RankingWeights
from snowflake.core.cortex.search_service._generated.models.result_row import ResultRow
from snowflake.core.cortex.search_service._generated.models.retrieval_weights import RetrievalWeights
from snowflake.core.cortex.search_service._generated.models.scheduling_state import SchedulingState
from snowflake.core.cortex.search_service._generated.models.scoring_config import ScoringConfig
from snowflake.core.cortex.search_service._generated.models.send_feedback_request import SendFeedbackRequest
from snowflake.core.cortex.search_service._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.cortex.search_service._generated.models.success_response import SuccessResponse
from snowflake.core.cortex.search_service._generated.models.suggest_query_request import SuggestQueryRequest
from snowflake.core.cortex.search_service._generated.models.suggest_query_response import SuggestQueryResponse
from snowflake.core.cortex.search_service._generated.models.suggest_result import SuggestResult
from snowflake.core.cortex.search_service._generated.models.target_lag import TargetLag
from snowflake.core.cortex.search_service._generated.models.text_boost import TextBoost
from snowflake.core.cortex.search_service._generated.models.time_decay import TimeDecay
from snowflake.core.cortex.search_service._generated.models.user_defined_lag import UserDefinedLag
from snowflake.core.cortex.search_service._generated.models.vector_boost import VectorBoost
from snowflake.core.cortex.search_service._generated.models.weights import Weights

__all__ = [
    "ColumnQuery",
    "CortexSearchService",
    "CreateCortexSearchServiceRequest",
    "Diversity",
    "DownstreamLag",
    "ErrorResponse",
    "Functions",
    "NumericBoost",
    "QueryRequest",
    "QueryResponse",
    "QuerySimilarityBoost",
    "RankingWeights",
    "ResultRow",
    "RetrievalWeights",
    "SchedulingState",
    "ScoringConfig",
    "SendFeedbackRequest",
    "SuccessAcceptedResponse",
    "SuccessResponse",
    "SuggestQueryRequest",
    "SuggestQueryResponse",
    "SuggestResult",
    "TargetLag",
    "TextBoost",
    "TimeDecay",
    "UserDefinedLag",
    "VectorBoost",
    "Weights",
]



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from snowflake.core.external_volume._generated.api.external_volume_api import ExternalVolumeApi

__all__ = [
    "ExternalVolumeApi",
]



"""Manages Snowflake artifact repository.

Example:
    >>> new_artifact_repository = ArtifactRepository(
    ...     name="my_artifact_repo", type="PIP", api_integration="my_api_integration"
    ... )
    >>> artifact_repositories = root.databases["MYDB"].schemas["MYSCHEMA"].artifact_repositories
    >>> my_artifact_repo = artifact_repositories.create(new_artifact_repository)
    >>> my_artifact_repo_snapshot = my_artifact_repo.fetch()
    >>> ar_data = artifact_repositories.iter(like="%my")
    >>> an_existing_repo = artifact_repositories["an_existing_repo"]
    >>> an_existing_repo.drop()

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from ._artifact_repository import ArtifactRepositoryCollection, ArtifactRepositoryResource
from ._generated.models.artifact_repository import ArtifactRepository


__all__ = ["ArtifactRepository", "ArtifactRepositoryCollection", "ArtifactRepositoryResource"]
